import logging

import numpy as np
from scipy.optimize import least_squares, minimize, fmin
from scipy.signal import find_peaks


logger = logging.getLogger(__name__)


class FitError(Exception):
    pass


def lorentz(x, w0, fwhm, intensity):
    with np.errstate(divide='ignore', invalid='ignore'):
        return intensity *\
               ((fwhm / 2) ** 2) / ((x - w0) ** 2 + (fwhm / 2) ** 2)


def fit_lorentz(x, y):
    w0_guess = float(x[np.argmax(y)])
    offset_guess = (y[0] + y[-1]) / 2.
    intensity_guess = np.max(y) - offset_guess
    fwhm_guess = float(2 * np.abs(
        w0_guess - x[np.argmax(y > (offset_guess + intensity_guess / 2))]))
    # Ensure a value larger than zero for the FWHM
    # (it might fail if the peak is not symmetric and its maximum is the
    # first value higher than offset_guess + intensity_guess/2)
    if fwhm_guess <= 0.0:
        fwhm_guess = 1.0

    def error(params, xdata, ydata):
        return (ydata
                - lorentz(xdata, *params[0:3])
                - params[3]) ** 2

    opt_result = least_squares(
        error,
        x0=(w0_guess, fwhm_guess, intensity_guess, offset_guess),
        args=(x, y)
    )

    if not opt_result.success:
        raise FitError('Lorentz fit failed.')

    w0, fwhm, intensity, offset = opt_result.x

    return w0, fwhm, intensity, offset


def fit_double_lorentz(x, y, bounds_w0=None):
    offset_guess = (y[0] + y[-1]) / 2.
    # gam_guess = (w0_guess ** 2 * (np.max(y) - offset_guess)) ** -0.5
    fwhm_guess = 4
    intensity_guess = np.max(y) - offset_guess

    # We run peak finding to get a good guess of the peak positions
    # and use the two peaks with highest prominence.
    peaks, properties = find_peaks(y, prominence=1)
    idx = np.argsort(properties['prominences'])[::-1]

    # Return if we didn't find two peaks
    if len(idx) < 2:
        return

    idx_sort = np.sort(peaks[idx[0:2]])
    w0_guess = list(x[idx_sort])

    def error(params, xdata, ydata):
        return (ydata
                - lorentz(xdata, *params[0:3])
                - lorentz(xdata, *params[3:6])
                - params[6]) ** 2

    # Create the bounds array
    if bounds_w0 is not None:
        # Lower limits
        bounds_lower = -np.Inf * np.ones(7)

        # 1st peak:
        # central position
        bounds_lower[0] = bounds_w0[0][0]
        # full-width-half-maximum
        # The VIPA spectrometer has an instrument width of
        # approx. 750 MHz, which translates to around
        # 7 px minimum peak width (FOB setup)
        # and 180 MHz and 4 px (780 nm setup).
        # We limit the minimal fit width to 3 px here.
        bounds_lower[1] = 3
        # intensity
        bounds_lower[2] = 0

        # 2nd peak:
        # central position
        bounds_lower[3] = bounds_w0[1][0]
        # full-width-half-maximum
        bounds_lower[4] = 3
        # intensity
        bounds_lower[5] = 0

        # offset
        bounds_lower[6] = 0

        # Upper limits
        bounds_upper = np.Inf * np.ones(7)

        bounds_upper[0] = bounds_w0[0][1]
        bounds_upper[3] = bounds_w0[1][1]
        bounds = (bounds_lower, bounds_upper)

        # Sort the guesses to the bounds
        bounds_w0_center = [np.mean(
            np.clip(bound, *x[::len(x) - 1])) for bound in bounds_w0]
        w0_guess.sort(reverse=(bounds_w0_center[0] > bounds_w0_center[1]))

        # Check that the initial guesses are within the bounds
        w0_guess = [np.clip(
            guess, *bounds_w0[idx]) for idx, guess in enumerate(w0_guess)]
    else:
        bounds = (-np.Inf, np.Inf)

    opt_result = least_squares(
        error,
        x0=(w0_guess[0], fwhm_guess, intensity_guess,
            w0_guess[1], fwhm_guess, intensity_guess,
            offset_guess
            ),
        args=(x, y),
        bounds=bounds
    )

    if not opt_result.success:
        raise FitError('Lorentz fit failed.')

    res = opt_result.x
    w0s, fwhms, intens = (res[0], res[3]), (res[1], res[4]), (res[2], res[5])
    offset = res[6]
    return w0s, fwhms, intens, offset


def fit_circle(points):
    """
    Fits a circle to a given set of points. Returnes the center and the radius
    of the cricle. The inital parameters for the fiting process are arbitrarily
    chosen.

    Parameters
    ----------
    points: list of tuples

    Returns
    -------
    center_opt: tuple
                Coordinates of the circle center (x,y)
    radius_opt: float
                Radius of the circle
    """

    center, radius = calculate_exact_circle(points)

    # No need to fit if we only have three points
    if len(points) <= 3:
        return center, radius

    param_guess = [center[0], center[1], radius]
    x_coords = np.array([xy[0] for xy in points])
    y_coords = np.array([xy[1] for xy in points])
    bnds = ((None, None), (None, None), (0.0, None))
    opt_result = minimize(_circle_opt,
                          param_guess,
                          args=(x_coords, y_coords),
                          bounds=bnds)

    return (opt_result['x'][0], opt_result['x'][1]), abs(opt_result['x'][2])


def _circle_opt(c, x_coord, y_coord):
    """
    Cost function to fit a circle to a given set of points.

    Parameters
    ----------
    c: array like
        Parameters to optimize: [x coordinates of circle center,
        y coordinates of circle center, radius of the cricle]
    x_coord:
        x coordinates of the points to fit
    y_coord
        y coordinates of the points to fit

    Returns
    -------
    out: float
        Cost value representing the deviation of the given set of points to
        the circle, which is represented by the parameter vector c.
    """
    return np.sum(
        ((x_coord - c[0]) ** 2
         + (y_coord - c[1]) ** 2
         - c[2] ** 2) ** 2)


def fit_lorentz_region(region, xdata, ydata, nr_peaks=1, bounds_w0=None):
    """
    Fits a lorentz or double lorentz fit to the given region

    Parameters
    ----------
    region: The section of the data to fit
    xdata: The x-data
    ydata: The y-data to fit
    nr_peaks: The number of peaks to fit
    bounds_w0: The bounds for the lorentz fit value of the maximum position

    Returns
    -------
    center, full-width-half-maximum, intensity and offset
    """
    try:
        if nr_peaks == 2:
            w0s, fwhms, intensities, offset = fit_double_lorentz(
                xdata[range(*region)], ydata[range(*region)],
                bounds_w0=bounds_w0)
        elif nr_peaks == 1:
            w0s, fwhms, intensities, offset = fit_lorentz(
                xdata[range(*region)], ydata[range(*region)])
        else:
            return
    except Exception:
        w0s = fwhms = intensities = offset = np.nan
    finally:
        return w0s, fwhms, intensities, offset


def calculate_exact_circle(points):
    # We need at least three points to fit a circle
    if len(points) < 3:
        return

    # If there are more than three points given,
    # we use the first and last and one in the center
    points = [points[idx]
              for idx in np.linspace(0, len(points) - 1, 3, dtype=int)]

    # If the first three points are all on the same line,
    # we slightly shift the first point so that we can
    # fit a circle
    if are_points_on_line(points):
        (x, y) = points[0]
        points[0] = (x - 1, y)

    x1 = points[0][0]
    y1 = points[0][1]
    x2 = points[1][0]
    y2 = points[1][1]
    x3 = points[2][0]
    y3 = points[2][1]

    x12 = x1 - x2
    x13 = x1 - x3

    y12 = y1 - y2
    y13 = y1 - y3

    sx13 = x1 ** 2 - x3 ** 2
    sy13 = y1 ** 2 - y3 ** 2

    sx21 = x2 ** 2 - x1 ** 2
    sy21 = y2 ** 2 - y1 ** 2

    x0 = -(sx13 * y12 +
           sy13 * y12 +
           sx21 * y13 +
           sy21 * y13) /\
        (2 * (x12 * y13 - x13 * y12))

    y0 = -(sx13 * x12 +
           sy13 * x12 +
           sx21 * x13 +
           sy21 * x13) /\
        (2 * (y12 * x13 - y13 * x12))

    c = (-x1 ** 2 - y1 ** 2 +
         2 * x0 * x1 + 2 * y0 * y1)

    r = np.sqrt(x0 ** 2 + y0 ** 2 - c)

    return (x0, y0), r


def are_points_on_line(points):
    """
    This function checks if the first three points in the given
    list lay on the same line
    Parameters
    ----------
    points: list of tuples representing points

    Returns
    -------
    boolean
    Whether the given points lay on the same line
    """
    if len(points) < 3:
        return True
    # We calculate a line form the first two points
    # and check whether the remaining points lay on it
    m = (points[0][1] - points[1][1]) /\
        (points[0][0] - points[1][0])
    n = points[0][1] - m * points[0][0]

    for point in points[:3]:
        if point[1] != (m * point[0] + n):
            return False

    return True


def fit_vipa(peaks, setup):
    """
    Fits the VIPA frequency axis

    Parameters
    ----------
    peaks: The peak positions
    setup: The setup parameters

    Returns
    -------
    The fitted VIPA parameters
    """
    # Check that we were given enough peaks
    if peaks is None:
        return
    if len(peaks) < (setup.calibration.num_brillouin_samples * 2 + 2):
        return

    # Calculate the start parameters for the VIPA fit
    r0 = peaks[0] * setup.pixel_size

    v1 = (setup.VIPA_PARAMS[0] +
          setup.VIPA_PARAMS[1] * r0 +
          setup.VIPA_PARAMS[2] * r0 ** 2) /\
         ((setup.vipa.m + setup.vipa.order) * np.pi)
    v2 = (setup.VIPA_PARAMS[1] +
          setup.VIPA_PARAMS[2] * 2 * r0) /\
         ((setup.vipa.m + setup.vipa.order) * np.pi) *\
        setup.pixel_size
    v3 = setup.VIPA_PARAMS[2] /\
        ((setup.vipa.m + setup.vipa.order) * np.pi) *\
        setup.pixel_size ** 2
    vipa_start = np.array([v1, v2, v3, setup.vipa.FSR])

    def error(vipa_params, peaks1, shifts1):
        fsr = vipa_params[3]
        frequencies = VIPA(peaks1, vipa_params) - setup.f0

        # Should match the expected frequencies
        d1 = frequencies - shifts1 - fsr * setup.calibration.orders

        # Should give equal values for Stokes and Anti-Stokes
        d2 = np.array([
            (frequencies[1] - frequencies[0]) -
            (frequencies[-1] - frequencies[-2]),
            (frequencies[2] - frequencies[0]) -
            (frequencies[-1] - frequencies[-3])
        ])

        # Should match the Brillouin shifts
        d3 = np.array([
            shifts1[1] - frequencies[1] + frequencies[0],
            shifts1[2] - frequencies[2] + frequencies[0],
            shifts1[2] - frequencies[5] + frequencies[3],
            shifts1[1] - frequencies[5] + frequencies[4]])

        return np.sum(d1 ** 2) + np.sum(d2 ** 2) + np.sum(d3 ** 2)

    opt_result = fmin(error,
                      vipa_start,
                      args=(peaks, setup.calibration.shifts))

    return opt_result[0], opt_result[1], opt_result[2], opt_result[3]


def VIPA(x, vipa_params):
    """
    Returns the absolute frequency in Hz of a given point
    on the spectrum in pixels.
    Subtract the absolute laser frequency from this value
    to get the relative shift.

    Parameters
    ----------
    x:   [pix]   The point on the spectrum
    vipa_params: Fit parameters of the VIPA fit

    Returns
    -------
    [Hz]   The frequency of the given point on the spectrum
    """
    return 1. / (vipa_params[0] + vipa_params[1] * x + vipa_params[2] * x ** 2)



from pathlib import Path
import pytest
import shutil
import uuid
import os

from bmlab.session import Session
from bmlab.controllers import ExportController


@pytest.fixture()
def tmp_dir():
    current_path = Path.cwd()
    tmp_dir = Path.cwd() / f"tmp{str(uuid.uuid4())}" / 'RawData'
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir, exist_ok=True)
    os.chdir(tmp_dir)
    yield tmp_dir
    os.chdir(current_path)
    try:
        shutil.rmtree(tmp_dir.parent)
    # Windows sometimes does not correctly close the HDF file. Then we have a
    # file access conflict.
    except Exception as e:
        print(e)


def data_file_path(file_name):
    return Path(__file__).parent / 'data' / file_name


def test_export_fluorescence(tmp_dir):
    shutil.copy(
        data_file_path('Fluorescence.h5'), Path.cwd() / 'Fluorescence.h5')

    session = Session.get_instance()
    session.set_file(Path('Fluorescence.h5'))

    ec = ExportController()
    config = ec.get_configuration()
    config['fluorescenceCombined']['export'] = False
    config['brillouin']['export'] = False
    ec.export(config)

    session.clear()

    plots_dir = tmp_dir.parent / 'Plots'
    images = [
        'Fluorescence_FLrep0_channelBlue.png',
        'Fluorescence_FLrep0_channelBlue_aligned.png',
        'Fluorescence_FLrep0_channelBrightfield.png',
        'Fluorescence_FLrep0_channelBrightfield_aligned.png',
        'Fluorescence_FLrep0_channelGreen.png',
        'Fluorescence_FLrep0_channelGreen_aligned.png',
        'Fluorescence_FLrep0_channelRed.png',
        'Fluorescence_FLrep0_channelRed_aligned.png',
        'Fluorescence_FLrep1_channelBlue.png',
        'Fluorescence_FLrep1_channelBlue_aligned.png',
        'Fluorescence_FLrep1_channelBrightfield.png',
        'Fluorescence_FLrep1_channelBrightfield_aligned.png',
        'Fluorescence_FLrep1_channelGreen.png',
        'Fluorescence_FLrep1_channelGreen_aligned.png',
        'Fluorescence_FLrep1_channelRed.png',
        'Fluorescence_FLrep1_channelRed_aligned.png',
    ]
    for image in images:
        assert os.path.exists(plots_dir / image)


def test_export_fluorescence_combined(tmp_dir):
    shutil.copy(
        data_file_path('Fluorescence.h5'), Path.cwd() / 'Fluorescence.h5')

    session = Session.get_instance()
    session.set_file(Path('Fluorescence.h5'))

    ec = ExportController()
    config = ec.get_configuration()
    config['fluorescence']['export'] = False
    config['brillouin']['export'] = False
    ec.export(config)

    session.clear()

    plots_dir = tmp_dir.parent / 'Plots' / 'Bare'
    images = [
        'Fluorescence_FLrep0_fluorescenceCombined___b.png',
        'Fluorescence_FLrep0_fluorescenceCombined___b_aligned.png',
        'Fluorescence_FLrep0_fluorescenceCombined__g_.png',
        'Fluorescence_FLrep0_fluorescenceCombined__g__aligned.png',
        'Fluorescence_FLrep0_fluorescenceCombined__gb.png',
        'Fluorescence_FLrep0_fluorescenceCombined__gb_aligned.png',
        'Fluorescence_FLrep0_fluorescenceCombined_r__.png',
        'Fluorescence_FLrep0_fluorescenceCombined_r___aligned.png',
        'Fluorescence_FLrep0_fluorescenceCombined_r_b.png',
        'Fluorescence_FLrep0_fluorescenceCombined_r_b_aligned.png',
        'Fluorescence_FLrep0_fluorescenceCombined_rg_.png',
        'Fluorescence_FLrep0_fluorescenceCombined_rg__aligned.png',
        'Fluorescence_FLrep0_fluorescenceCombined_rgb.png',
        'Fluorescence_FLrep0_fluorescenceCombined_rgb_aligned.png',
        'Fluorescence_FLrep1_fluorescenceCombined___b.png',
        'Fluorescence_FLrep1_fluorescenceCombined___b_aligned.png',
        'Fluorescence_FLrep1_fluorescenceCombined__g_.png',
        'Fluorescence_FLrep1_fluorescenceCombined__g__aligned.png',
        'Fluorescence_FLrep1_fluorescenceCombined__gb.png',
        'Fluorescence_FLrep1_fluorescenceCombined__gb_aligned.png',
        'Fluorescence_FLrep1_fluorescenceCombined_r__.png',
        'Fluorescence_FLrep1_fluorescenceCombined_r___aligned.png',
        'Fluorescence_FLrep1_fluorescenceCombined_r_b.png',
        'Fluorescence_FLrep1_fluorescenceCombined_r_b_aligned.png',
        'Fluorescence_FLrep1_fluorescenceCombined_rg_.png',
        'Fluorescence_FLrep1_fluorescenceCombined_rg__aligned.png',
        'Fluorescence_FLrep1_fluorescenceCombined_rgb.png',
        'Fluorescence_FLrep1_fluorescenceCombined_rgb_aligned.png',
    ]
    for image in images:
        assert os.path.exists(plots_dir / image)


def test_export_brillouin(tmp_dir):
    shutil.copy(
        data_file_path('2D-xy.h5'), Path.cwd() / '2D-xy.h5')

    session = Session.get_instance()
    session.set_file(Path('2D-xy.h5'))

    ec = ExportController()
    config = ec.get_configuration()
    config['fluorescence']['export'] = False
    config['fluorescenceCombined']['export'] = False
    config['brillouin']['parameters'] =\
        ['brillouin_shift_f', 'brillouin_shift']
    ec.export(config)

    session.clear()

    plots_dir = tmp_dir.parent / 'Plots' / 'Bare'
    images = [
        '2D-xy_BMrep0_brillouin_shift_f.png',
        '2D-xy_BMrep0_brillouin_shift_f.tiff',
        '2D-xy_BMrep0_brillouin_shift.png',
        '2D-xy_BMrep0_brillouin_shift.tiff',
    ]
    for image in images:
        assert os.path.exists(plots_dir / image)

    plots_dir = tmp_dir.parent / 'Plots' / 'WithAxis'
    images = [
        '2D-xy_BMrep0_brillouin_shift_f.pdf',
        '2D-xy_BMrep0_brillouin_shift_f.png',
        '2D-xy_BMrep0_brillouin_shift.pdf',
        '2D-xy_BMrep0_brillouin_shift.png',
    ]
    for image in images:
        assert os.path.exists(plots_dir / image)

    csvs = [
        '2D-xy_BMrep0_brillouin_shift_f.csv',
        '2D-xy_BMrep0_brillouin_shift.csv',
    ]
    for csv in csvs:
        assert os.path.exists(
            tmp_dir.parent / 'Export' / csv)



import pathlib

from bmlab.session import Session
from bmlab.controllers import ExtractionController, CalibrationController
from bmlab.models import Orientation
from bmlab.models.setup import AVAILABLE_SETUPS


def test_find_peaks_real_data():
    # Start session
    session = Session.get_instance()

    # Load data file
    session.set_file(pathlib.Path(__file__).parent / 'data' / 'Water.h5')

    # Select repetition
    session.set_current_repetition('0')
    session.set_setup(AVAILABLE_SETUPS[0])

    # Set orientation
    session.orientation = Orientation(rotation=1, reflection={
        'vertically': False, 'horizontally': False
    })

    cm = session.calibration_model()

    ec = ExtractionController()
    cc = CalibrationController()

    # First add all extraction points because this
    # can influence the extraction for other calibrations
    for calib_key in session.get_calib_keys():
        ec.find_points(calib_key)

    brillouin_region_centers = [230, 330]
    rayleigh_region_centers = [130, 390]

    # Then do the calibration
    for calib_key in session.get_calib_keys():
        cc.find_peaks(calib_key)

        brillouin_regions = cm.get_brillouin_regions(calib_key)
        rayleigh_regions = cm.get_rayleigh_regions(calib_key)

        for center in brillouin_region_centers:
            assert region_found(center, brillouin_regions)

        for center in rayleigh_region_centers:
            assert region_found(center, rayleigh_regions)


def region_found(center, regions):
    for region in regions:
        if region[0] <= center <= region[1]:
            return True
    return False


def test_calibrate():
    calib_key = '0'
    cc = CalibrationController()
    cc.calibrate(calib_key)



import pytest
import pathlib
import datetime

from bmlab.file import BrillouinFile,\
    BadFileException, is_source_file, is_session_file


def data_file_path(file_name):
    return pathlib.Path(__file__).parent / 'data' / file_name


def test_is_source_file():
    # File from BrillouinAcquisition
    assert is_source_file(data_file_path('Water.h5'))
    # Session file from bmlab<0.0.14
    assert not is_source_file(data_file_path('1D-x.session.h5'))
    # Session file from bmlab>=0.0.14
    assert not is_source_file(data_file_path('1D-y.session.h5'))
    # Non-existing file
    assert not is_source_file(data_file_path('Unavailable.h5'))


def test_is_session_file():
    # Session file from bmlab<0.0.14
    assert is_session_file(data_file_path('1D-x.session.h5'))
    # Session file from bmlab>=0.0.14
    assert is_session_file(data_file_path('1D-y.session.h5'))
    # File from BrillouinAcquisition
    assert not is_session_file(data_file_path('Water.h5'))
    # Non-existing file
    assert not is_session_file(data_file_path('Unavailable.h5'))


def test_file_has_one_repetition():
    bf = BrillouinFile(data_file_path('Water.h5'))
    num_repetitions = bf.repetition_count()
    assert num_repetitions == 1

    bf = BrillouinFile(data_file_path('Water_old.h5'))
    num_repetitions = bf.repetition_count()
    assert num_repetitions == 1


def test_non_existing_file_raises_exception():
    with pytest.raises(OSError):
        BrillouinFile(data_file_path('non_existing_file.h5'))


def test_file_has_comment():
    bf = BrillouinFile(data_file_path('Water.h5'))
    assert bf.comment == 'Brillouin data'


def test_file_has_date():
    bf = BrillouinFile(data_file_path('Water.h5'))
    assert bf.date == datetime.datetime.fromisoformat(
        '2020-11-03T15:20:30.682+01:00')


def test_open_file_with_no_brillouin_data_raises_exception():
    with pytest.raises(BadFileException):
        BrillouinFile(data_file_path('empty_file.h5'))


def test_file_get_repetition_keys():
    bf = BrillouinFile(data_file_path('Water.h5'))
    assert bf.repetition_keys() == ['0']

    assert bf.repetition_keys('Fluorescence') == []

    bf = BrillouinFile(data_file_path('Water_old.h5'))
    assert bf.repetition_keys() == ['0']

    assert bf.repetition_keys('Fluorescence') == []


def test_file_get_repetition_date():
    bf = BrillouinFile(data_file_path('Water.h5'))
    rep = bf.get_repetition('0')
    assert rep.date == datetime.datetime.fromisoformat(
        '2020-11-03T15:20:52.852+01:00')


def test_file_get_resolution():
    bf = BrillouinFile(data_file_path('Water.h5'))
    rep = bf.get_repetition('0')
    assert rep.payload.resolution == (10, 1, 1)


def test_file_get_positions():
    bf = BrillouinFile(data_file_path('Water.h5'))
    rep = bf.get_repetition('0')
    # Test that the shape is correct
    assert rep.payload.positions['x'].shape == (1, 10, 1)
    assert rep.payload.positions['y'].shape == (1, 10, 1)
    assert rep.payload.positions['z'].shape == (1, 10, 1)
    # Test some values
    assert rep.payload.positions['x'][0, 0, 0] == -5652.5
    assert rep.payload.positions['y'][0, 0, 0] == -1963.0
    assert rep.payload.positions['z'][0, 0, 0] == 201.425


def test_file_repetition_has_calibration():
    bf = BrillouinFile(data_file_path('Water.h5'))
    rep = bf.get_repetition('0')
    assert not rep.calibration.is_empty()

    bf = BrillouinFile(data_file_path('Water_old.h5'))
    rep = bf.get_repetition('0')
    assert not rep.calibration.is_empty()


def test_file_payload_image_keys():
    bf = BrillouinFile(data_file_path('Water.h5'))
    rep = bf.get_repetition('0')
    assert rep.payload.image_keys() == [str(k) for k in range(10)]

    bf = BrillouinFile(data_file_path('Water_old.h5'))
    rep = bf.get_repetition('0')
    assert rep.payload.image_keys() == [str(k) for k in range(4)]


def test_file_payload_get_image():
    bf = BrillouinFile(data_file_path('Water.h5'))
    image = bf.get_repetition('0').payload.get_image('0')
    assert image.shape == (2, 400, 400)

    bf = BrillouinFile(data_file_path('Water_old.h5'))
    image = bf.get_repetition('0').payload.get_image('0')
    assert image.shape == (2, 400, 400)


def test_file_payload_get_date():
    bf = BrillouinFile(data_file_path('Water.h5'))
    date = bf.get_repetition('0').payload.get_date('0')
    assert date == datetime.datetime.fromisoformat(
        '2020-11-03T15:21:10.568+01:00')


def test_file_payload_get_time():
    bf = BrillouinFile(data_file_path('Water.h5'))
    time = bf.get_repetition('0').payload.get_time('0')
    assert time == 39.886


def test_file_calibration_image_keys():
    bf = BrillouinFile(data_file_path('Water.h5'))
    rep = bf.get_repetition('0')
    assert rep.calibration.image_keys() == [str(k+1) for k in range(2)]

    bf = BrillouinFile(data_file_path('Water_old.h5'))
    rep = bf.get_repetition('0')
    assert rep.calibration.image_keys() == [str(k+1) for k in range(2)]


def test_file_calibration_get_date():
    bf = BrillouinFile(data_file_path('Water.h5'))
    date = bf.get_repetition('0').calibration.get_date('1')
    assert date == datetime.datetime.fromisoformat(
        '2020-11-03T15:21:07.484+01:00')


def test_file_calibration_get_time():
    bf = BrillouinFile(data_file_path('Water.h5'))
    time = bf.get_repetition('0').calibration.get_time('1')
    assert time == 36.802


def test_file_calibration_get_exposure():
    bf = BrillouinFile(data_file_path('Binning.h5'))
    assert bf.get_repetition('0').calibration.get_exposure('1') == 0.8
    assert bf.get_repetition('0').payload.get_exposure('0') == 0.2


def test_file_calibration_get_binning():
    bf = BrillouinFile(data_file_path('Binning.h5'))
    assert bf.get_repetition('0').calibration.get_binning('1') == '8x8'
    assert bf.get_repetition('0').payload.get_binning('0') == '8x8'

    assert bf.get_repetition('0')\
        .payload.get_binning_factor('0') == 8


def test_file_repetition_count():
    bf = BrillouinFile(data_file_path('Fluorescence.h5'))
    mode = 'Fluorescence'

    rep_keys = bf.repetition_keys(mode)
    assert rep_keys == ['0', '1']

    with pytest.raises(NotImplementedError):
        bf.repetition_count('Raman')


def test_file_get_fluorescence_images():
    bf = BrillouinFile(data_file_path('Fluorescence.h5'))
    mode = 'Fluorescence'

    rep_keys = bf.repetition_keys(mode)
    assert rep_keys == ['0', '1']

    repetition = bf.get_repetition(rep_keys[0], mode)

    image_keys = repetition.payload.image_keys()
    assert image_keys == ['0', '1', '2', '3']

    time = repetition.payload.get_time(image_keys[0])
    assert time == 22.376

    channel = repetition.payload.get_channel(image_keys[0])
    assert channel == 'Blue'

    roi = repetition.payload.get_ROI(image_keys[0])
    assert roi == dict({
        'bottom': 152,
        'height_binned': 700,
        'height_physical': 700,
        'left': 200,
        'right': 202,
        'top': 150,
        'width_binned': 600,
        'width_physical': 600
    })

    image = repetition.payload.get_image(image_keys[0])
    assert image.shape == (1, 700, 600)


def test_file_get_scale_calibration():
    bf = BrillouinFile(data_file_path('Fluorescence.h5'))
    mode = 'Fluorescence'

    repetition = bf.get_repetition('0', mode)
    scale_calibration = repetition.payload.get_scale_calibration()

    assert scale_calibration == dict({
        'micrometerToPixX': (-7.95, 9.05),
        'micrometerToPixY': (9.45, 8.65),
        'pixToMicrometerX': (-0.05606325750210643, 0.05865577808023852),
        'pixToMicrometerY': (0.0612482986583706, 0.05152634649037527),
        'positionScanner': (0.0, 0.0),
        'positionStage': (0.0, 0.0),
        'origin': (0.0, 0.0)
    })



"""
Module for interacting with files containing Brillouin microscopy
data.

NOTE: Ideally, the users of bmlab should not have to know about
the file format in which the data are stored. So, if possible,
do not expose HDF objects to the outside (like BMicro).
"""

import datetime

import numpy as np
import h5py
from pathlib import Path

from packaging import version

BRILLOUIN_GROUP = 'Brillouin'
FLUORESCENCE_GROUP = 'Fluorescence'


def _get_datetime(time_stamp):
    """ Convert the time stamp in the HDF file to Python datetime """
    try:
        return datetime.datetime.fromisoformat(time_stamp)
    except Exception:
        return None


def is_source_file(path):
    try:
        with h5py.File(path, "r") as h5:
            file_format = h5.attrs.get('version')
            if file_format is not None:
                # Is this a file from BrillouinAcquisition?
                # Then we get an ndarray with a bytes object inside
                # (because the version attribute is stored
                # as 'cset=H5T_CSET_ASCII').
                if isinstance(file_format, np.ndarray)\
                        and isinstance(file_format[0], bytes):
                    file_format = file_format[0].decode('ascii')
                # This is from BrillouinAcquisition!
                if file_format.startswith('H5BM'):
                    return True
        return False
    except Exception:
        return False


def is_session_file(path):
    try:
        path = Path(path)
        with h5py.File(path, "r") as h5:
            file_format = h5.attrs.get('version')
            if file_format is not None:
                # This is a bmlab session file created
                # with bmlab>=0.0.14
                if isinstance(file_format, str)\
                        and file_format.startswith('bmlab'):
                    return True

            # This is a bmlab session file created with bmlab<0.0.14
            if 'session' in h5.keys()\
                    and h5['session'].attrs.get('type').startswith('bmlab'):
                return True
        return False
    except Exception:
        return False


class BrillouinFile(object):

    def __init__(self, path):
        """
        Load a HDF file with Brillouin microscopy data.

        Parameters
        ----------
        path : Path
            path of the file to load

        Raises
        ------
        OSError
            when trying to open non-existing or bad file
        """
        self.path = Path(path).resolve()
        self.file = None
        self.file = h5py.File(self.path, 'r')
        self.file_version_string = self.file.attrs.get('version')[
            0].decode('utf-8')
        if not self.file_version_string.startswith('H5BM'):
            raise BadFileException('File does not contain any Brillouin data')
        self.file_version = self.file_version_string[-5:]
        # Fluorescence group is optional
        self.Fluorescence_group = None

        if version.parse(self.file_version) >= version.parse("0.0.4"):
            """"
            New Brillouin file format,
            supporting different modes and repetitions
            """
            if BRILLOUIN_GROUP not in self.file:
                raise BadFileException(
                    'File does not contain any Brillouin data')
            self.Brillouin_group = self.file[BRILLOUIN_GROUP]

            if FLUORESCENCE_GROUP in self.file:
                self.Fluorescence_group = self.file[FLUORESCENCE_GROUP]
        else:
            """ Old Brillouin file format """
            self.Brillouin_group = self.file
        # Comments are optional,
        # e.g. for files containing only fluorescence data
        comment = self.file.attrs.get('comment')
        if comment is not None:
            self.comment = comment[0].decode('utf-8')
        self.date = _get_datetime(
            self.file.attrs.get('date')[0].decode('utf-8'))

    def __del__(self):
        """
        Destructor. Closes hdf file when object runs out of scope.
        """
        self.close()

    def close(self):
        try:
            self.file.close()
        except Exception:
            pass

    @staticmethod
    def checkMode(mode):
        modes = [BRILLOUIN_GROUP, FLUORESCENCE_GROUP]
        if mode not in modes:
            raise NotImplementedError(
                'The mode "{}" is not supported.'.format(mode))

    def repetition_count(self, mode=BRILLOUIN_GROUP):
        """
        Get the number of repetitions in the data file.

        Returns
        -------
        out : int
            Number of repetitions in the data file
        """
        self.checkMode(mode)
        return len(self.repetition_keys(mode))

    def repetition_keys(self, mode=BRILLOUIN_GROUP):
        """
        Returns list of keys for the various repetitions in the file.

        Returns
        -------
        out: list of str
        """
        self.checkMode(mode)
        mode_group = self.Brillouin_group
        if mode == FLUORESCENCE_GROUP:
            mode_group = self.Fluorescence_group
        if mode_group is None:
            return []
        if version.parse(self.file_version) >= version.parse("0.0.4"):
            return list(mode_group.keys())
        elif mode == BRILLOUIN_GROUP:
            return list(['0'])
        else:
            return []

    def get_repetition(self, repetition_key, mode=BRILLOUIN_GROUP):
        """
        Get a repetition from the data file based on given key.

        Parameters
        ----------
        repetition_key : str
            key to identify the repetition in the Brillouin group
        mode: str
            the mode to look at

        Returns
        -------
        out : Repetition
            the repetition
        """
        self.checkMode(mode)
        mode_group = self.Brillouin_group
        if mode == FLUORESCENCE_GROUP:
            mode_group = self.Fluorescence_group
        # Check that we have a group for this mode
        if mode_group is None:
            return None
        if version.parse(self.file_version) >= version.parse("0.0.4"):
            return Repetition(mode_group.get(repetition_key), self)
        else:
            return Repetition(self.file, self)


class Repetition(object):

    def __init__(self, repetition_group, file):
        """
        Creates a repetition from the corresponding group of a HDF file.

        Parameters
        ----------
        repetition_group : HDF group
            The HDF group representing a Repetition. Consists of payload,
            calibration and background.
        """
        self.date = _get_datetime(
            repetition_group.attrs.get('date')[0].decode('utf-8'))
        self.payload = Payload(repetition_group.get('payload'), self)
        calibration_group = repetition_group.get('calibration')
        # Having a calibration is optional for a repetition
        if calibration_group is not None:
            self.calibration = Calibration(calibration_group, self)
        self.file = file


class MeasurementData(object):

    def __init__(self, payload_group, repetition):
        """
        Creates a payload representation from the corresponding group of a
        HDF file.

        Parameters
        ----------
        payload_group : HDF group
            The payload of a repetition, basically a set of images

        """
        self.repetition = repetition
        self.group = payload_group
        if payload_group is not None:
            self.data = payload_group.get('data')
        else:
            self.data = None

    def image_keys(self, sort_by_time=False):
        """
        Returns the keys of the images stored in the payload,
        optionally sorted by time.

        Parameters
        ----------
        sort_by_time : bool

        Returns
        -------
        out: list of str
            Keys of images in payload.
        """
        if self.data:
            keys = list(self.data.keys())
            if not sort_by_time:
                return keys

            dates = []
            for key in keys:
                dates.append(self.get_date(key))

            return [i for _, i in sorted(zip(dates, keys))]
        return []

    def get_image(self, image_key):
        """
        Returns the image from the calibration for given key.

        Parameters
        ----------
        image_key: str
            Key for the image.

        Returns
        -------
        out: numpy.ndarray
            Array representing the image.
        """
        imgs = self.data.get(image_key)
        if imgs is None:
            return None
        return np.array(imgs)

    def get_image_count(self, image_key):
        imgs = self.data.get(image_key)
        if imgs is None:
            return 0
        return imgs.shape[0]

    def get_date(self, image_key):
        """"
        Returns the date of an image
        with the given key
        """
        try:
            return _get_datetime(
                self.data.get(image_key).attrs.get('date')[0].decode('utf-8'))
        except Exception:
            return ''

    def get_time(self, image_key):
        try:
            # Get date of the calibration
            date = self.get_date(image_key)
            # Get the reference date
            ref = self.repetition.file.date
            # return the difference in seconds
            return (date - ref).total_seconds()
        except Exception:
            return None

    def is_empty(self):
        return self.data is None or len(self.data) == 0

    def get_exposure(self, image_key):
        """"
        Returns the exposure time of a payload image
        with the given key
        """
        try:
            return self.data.get(image_key).attrs\
                .get('exposure')[0]
        except Exception:
            # For older files we return a default value
            return 0.5

    def get_binning(self, image_key):
        """"
        Returns the binning of a payload image
        with the given key
        """
        try:
            return self.data.get(image_key).attrs\
                .get('binning')[0].decode('utf-8')
        except Exception:
            # For older files we return a default value
            return '1x1'

    def get_binning_factor(self, image_key):
        binning = self.get_binning(image_key)
        # We only check the values possible
        # with BrillouinAcquisition
        if binning == '2x2':
            return 2
        elif binning == '4x4':
            return 4
        elif binning == '8x8':
            return 8
        return 1

    def get_channel(self, image_key):
        """"
        Returns the channel of a payload image
        with the given key
        """
        try:
            return self.data.get(image_key).attrs\
                .get('channel')[0].decode('utf-8')
        except Exception:
            return None

    def get_ROI(self, image_key):
        """
        Returns the region of interest of a payload image
        with the given key
        Parameters
        ----------
        image_key

        Returns
        -------

        """
        # Older measurements didn't save this information,
        # so we wrap it in a try/except
        try:
            attributes = ['left', 'right',
                          'bottom', 'top',
                          'width_physical', 'height_physical',
                          'width_binned', 'height_binned']
            roi = dict()
            for attribute in attributes:
                roi[attribute] =\
                    self.data.get(image_key).attrs.get('ROI_' + attribute)[0]

            return roi
        except BaseException:
            return None


class Payload(MeasurementData):

    def __init__(self, payload_group, repetition):
        """
        Creates a payload representation from the corresponding group of a
        HDF file.

        Parameters
        ----------
        payload_group : HDF group
            The payload of a repetition, basically a set of images

        """
        super(Payload, self).__init__(payload_group, repetition)
        # Only Brillouin payloads have a resolution and positions
        try:
            self.resolution = tuple(int(payload_group.attrs.get(
                'resolution-%s' % axis)[0]) for axis in ['x', 'y', 'z'])
            self.positions = {
                'x': np.array(payload_group.get('positions-x')),
                'y': np.array(payload_group.get('positions-y')),
                'z': np.array(payload_group.get('positions-z')),
            }
        except BaseException:
            self.resolution = None
            self.positions = None

    def get_scale_calibration(self):
        parameters = [
            'micrometerToPixX', 'micrometerToPixY',
            'pixToMicrometerX', 'pixToMicrometerY',
            'positionScanner', 'positionStage', 'origin'
        ]
        try:
            cal = self.group.get('scaleCalibration')
            scaleCal = dict()
            for attribute in parameters:
                val = cal.get(attribute)
                scaleCal[attribute] = \
                    tuple(val.attrs.get(dim)[0] for dim in ['x', 'y'])

            return scaleCal
        except BaseException:
            return None


class Calibration(MeasurementData):

    def __init__(self, payload_group, repetition):
        """
        Creates a calibration representation from the corresponding group of
        a HDF file.

        Parameters
        ----------
        payload_group : HDF group
            Calibration data of a repetition from an HDF file.
        """
        super(Calibration, self).__init__(payload_group, repetition)
        """
        For H5BM files < 0.0.4
        there was an inconsistency with the group naming
        """
        if self.data is None and payload_group is not None:
            self.data = payload_group.get('calibrationData')


class BadFileException(Exception):
    pass



import logging

import numpy as np
from scipy.signal import medfilt2d, find_peaks
from skimage.measure import label
from skimage.morphology import closing, disk

from math import floor

import multiprocessing as mp
from itertools import repeat as irepeat

from bmlab import Session
from bmlab.fits import fit_vipa, VIPA, fit_lorentz_region
from bmlab.image import extract_lines_along_arc, find_max_in_radius
from bmlab.export import FluorescenceExport,\
    FluorescenceCombinedExport, BrillouinExport

import warnings

logger = logging.getLogger(__name__)


class ExtractionController(object):

    def __init__(self):
        self.session = Session.get_instance()
        return

    def add_point(self, calib_key, point):
        time = self.session.get_calibration_time(calib_key)
        if time is None:
            return
        em = self.session.extraction_model()
        em.add_point(calib_key, time, *point)

    def set_point(self, calib_key, index, point):
        time = self.session.get_calibration_time(calib_key)
        if time is None:
            return
        em = self.session.extraction_model()
        em.set_point(calib_key, index, time, *point)

    def optimize_points(self, calib_key, radius=10):
        em = self.session.extraction_model()

        imgs = self.session.get_calibration_image(calib_key)
        if imgs is None:
            return
        img = np.nanmean(imgs, axis=0)
        img = medfilt2d(img)

        points = em.get_points(calib_key)
        time = em.get_time(calib_key)
        em.clear_points(calib_key)

        for p in points:
            new_point = find_max_in_radius(img, p, radius)
            # Warning: x-axis in imshow is 1-axis in img, y-axis is 0-axis
            em.add_point(
                calib_key, time, new_point[0], new_point[1])

    def find_points_all(self):
        calib_keys = self.session.get_calib_keys()

        if not calib_keys:
            return

        for calib_key in calib_keys:
            self.find_points(calib_key)

    def find_points(self, calib_key, min_height=10,
                    min_area=20, max_distance=50):
        em = self.session.extraction_model()

        imgs = self.session.get_calibration_image(calib_key)
        if imgs is None:
            return
        img = np.nanmean(imgs, axis=0)
        time = self.session.get_calibration_time(calib_key)
        if time is None:
            return

        # Account for binning for default values
        disc_size = 10
        binning = self.session.get_calibration_binning_factor(calib_key)
        if binning:
            max_distance = max_distance / binning
            disc_size = disc_size / binning
            min_area = min_area / (binning**2)

        img = medfilt2d(img)
        # This is the background level
        threshold = np.median(img)

        # Try to find a signal dependent estimate
        # for the minimal peak height.
        # Discard all values smaller than the background noise
        # plus a minimal peak height
        img_peaks = img[img > threshold + min_height]
        # Calculate the signal dependent peak threshold
        height = (np.nanmean(img_peaks) + threshold) / 2
        img_closed = closing(img > height, disk(disc_size))

        # Find all peaks higher than the min_height
        image_label, num = label(img_closed, return_num=True)

        all_peaks = []
        for region in range(1, num + 1):
            # Mask of everything but the peak
            mask = (image_label != region)
            # Set everything but the peak to zero
            tmp = img.copy()
            tmp[mask] = 0

            # Discard all peaks with an area that is too small
            if np.sum(np.logical_not(mask)) >= min_area:
                # Find indices of peak maximum
                ind = np.unravel_index(np.argmax(tmp, axis=None), tmp.shape)

                all_peaks.append(ind)

        # Filter found peaks
        p0 = (0, img.shape[1])
        p1 = (img.shape[0], 0)
        peaks = filter(
            lambda peak:
            self.distance_point_to_line(peak, p0, p1) < max_distance,
            all_peaks)

        # Add found peaks to model
        em.set_points(calib_key, time, list(peaks))

    def distance_point_to_line(self, point, line0, line1):
        return abs(
            (line1[1] - line0[1]) * (line0[0] - point[0]) -
            (line0[1] - point[1]) * (line1[0] - line0[0]))\
               / np.sqrt((line1[1] - line0[1])**2 + (line1[0] - line0[0])**2)


class ImageController(object):

    def __init__(self, model, get_image, get_time, get_exposure):
        self.session = Session.get_instance()
        self.model = model
        self.get_image = get_image
        self.get_time = get_time
        self.get_exposure = get_exposure

    def extract_spectra(self, image_key, frame_num=None):
        em = self.session.extraction_model()
        if not em:
            return None, None, None
        time = self.get_time(image_key)
        arc = em.get_arc_by_time(time)
        if arc.size == 0:
            return None, None, None

        imgs = self.get_image(image_key)
        if frame_num is not None:
            imgs = imgs[frame_num:frame_num+1]

        # Extract values from *all* frames in the current calibration
        spectra = []
        for img in imgs:
            values_by_img = extract_lines_along_arc(
                img,
                arc
            )
            spectra.append(values_by_img)

        exposure = self.get_exposure(image_key)
        times = exposure * np.arange(len(imgs)) + time

        intensities = np.nanmean(imgs, axis=(1, 2))

        # We only set the spectra if we extracted all
        if frame_num is None:
            self.model().set_spectra(image_key, spectra)
        return spectra, times, intensities


class CalibrationController(ImageController):

    def __init__(self, *args, **kwargs):
        session = Session.get_instance()
        super(CalibrationController, self).__init__(
            model=session.calibration_model,
            get_image=session.get_calibration_image,
            get_time=session.get_calibration_time,
            get_exposure=session.get_calibration_exposure
        )
        return

    def find_peaks(self, calib_key, min_prominence=15,
                   num_brillouin_samples=2, min_height=15):
        spectra, _, _ = self.extract_spectra(calib_key)
        if spectra is None:
            return
        spectrum = np.mean(spectra, axis=0)
        # This is the background value
        base = np.nanmedian(spectrum)
        peaks, properties = find_peaks(
            spectrum, prominence=min_prominence, width=True,
            height=min_height+base)

        # Number of peaks we are searching for
        # (2 Rayleigh + 2 times number calibration samples)
        num_peaks = 2 + 2 * num_brillouin_samples

        # Check if we found enough peak candidates
        if len(peaks) < num_peaks:
            # If we didn't find enough peaks, we try again
            # without a minimum height
            peaks, properties = find_peaks(
                spectrum, prominence=min_prominence, width=True)
            # If there a still too few, we give up
            if len(peaks) < num_peaks:
                return

        # We need to identify the position between the
        # Stokes and Anti-Stokes Brillouin peaks

        # In case there are just enough peaks,
        # we use the position in the middle:
        if len(peaks) == num_peaks:
            idx = int(num_peaks / 2)
            center = np.mean(peaks[idx - 1:idx + 1])
        # Otherwise we use the center of mass as the middle
        else:
            # Set everything below the background value to zero,
            # so it does not affect the center calculation
            spectrum[spectrum < base] = 0
            # Calculate the center of mass
            center = np.nansum(spectrum * range(1, len(spectrum) + 1))\
                / np.nansum(spectrum)

            # Check that we have enough peaks on both sides of the center
            num_peaks_right = len(peaks[peaks > center])
            num_peaks_left = len(peaks[peaks <= center])

            # If not enough peaks on the right, shift center to the left
            if num_peaks_right < (num_brillouin_samples + 1):
                center = np.mean(
                    peaks[-(num_brillouin_samples + 2):-num_brillouin_samples]
                )
            # If not enough peaks on the left, shift center to the right
            elif num_peaks_left < (num_brillouin_samples + 1):
                center = np.mean(
                    peaks[num_brillouin_samples:num_brillouin_samples + 2]
                )

        num_peaks_left = len(peaks[peaks <= center])

        indices_brillouin = range(
            num_peaks_left - num_brillouin_samples,
            num_peaks_left + num_brillouin_samples
        )
        indices_rayleigh = [
            num_peaks_left - num_brillouin_samples - 1,
            num_peaks_left + num_brillouin_samples
        ]

        def peak_to_region(i):
            r = (
                        peaks[i]
                        + properties['widths'][i] * np.array((-4, 4))
                ).astype(int)
            r[r > len(spectrum)] = len(spectrum)
            return tuple(r)

        regions_brillouin = list(map(peak_to_region, indices_brillouin))
        # Merge the Brillouin regions if necessary
        if num_brillouin_samples > 1:
            regions_brillouin = [
                (regions_brillouin[0][0],
                 regions_brillouin[num_brillouin_samples - 1][1]),
                (regions_brillouin[num_brillouin_samples][0],
                 regions_brillouin[-1][1]),
            ]

        regions_rayleigh = map(peak_to_region, indices_rayleigh)

        cm = self.session.calibration_model()
        # Add Brillouin regions
        for i, region in enumerate(regions_brillouin):
            # We use "set_brillouin_region" here so overlapping
            # regions don't get merged
            cm.set_brillouin_region(calib_key, i, region)
        # Add Rayleigh regions
        for i, region in enumerate(regions_rayleigh):
            # We use "set_brillouin_region" here so overlapping
            # regions don't get merged
            cm.set_rayleigh_region(calib_key, i, region)

    def calibrate(self, calib_key, count=None, max_count=None):

        setup = self.session.setup
        em = self.session.extraction_model()
        cm = self.session.calibration_model()

        if not calib_key\
                or not setup\
                or not em\
                or not cm:
            if max_count is not None:
                max_count.value = -1
            return

        spectra, _, _ = self.extract_spectra(calib_key)
        time = self.session.get_calibration_time(calib_key)

        if spectra is None or len(spectra) == 0:
            if max_count is not None:
                max_count.value = -1
            return

        self.fit_rayleigh_regions(calib_key)
        self.fit_brillouin_regions(calib_key)

        vipa_params = []
        frequencies = []

        if max_count is not None:
            max_count.value += len(spectra)

        for frame_num, spectrum in enumerate(spectra):
            peaks = cm.get_sorted_peaks(calib_key, frame_num)

            params = fit_vipa(peaks, setup)
            if params is None:
                continue
            vipa_params.append(params)
            xdata = np.arange(len(spectrum))

            frequencies.append(VIPA(xdata, params) - setup.f0)
            if count is not None:
                count.value += 1

        cm.set_vipa_params(calib_key, vipa_params)
        cm.set_frequencies(calib_key, time, frequencies)

        calculate_derived_values()

    def clear_calibration(self, calib_key):
        cm = self.session.calibration_model()
        if not cm:
            return

        cm.clear_brillouin_fits(calib_key)
        cm.clear_rayleigh_fits(calib_key)
        cm.clear_frequencies(calib_key)
        cm.clear_vipa_params(calib_key)

        calculate_derived_values()

    def fit_rayleigh_regions(self, calib_key):
        cm = self.session.calibration_model()
        spectra = cm.get_spectra(calib_key)
        regions = cm.get_rayleigh_regions(calib_key)

        cm.clear_rayleigh_fits(calib_key)
        for frame_num, spectrum in enumerate(spectra):
            for region_key, region in enumerate(regions):
                xdata = np.arange(len(spectrum))
                w0, fwhm, intensity, offset = \
                    fit_lorentz_region(region, xdata, spectrum)
                cm.add_rayleigh_fit(calib_key, region_key, frame_num,
                                    w0, fwhm, intensity, offset)

    def fit_brillouin_regions(self, calib_key):
        cm = self.session.calibration_model()
        spectra = cm.get_spectra(calib_key)
        regions = cm.get_brillouin_regions(calib_key)
        setup = self.session.setup
        if not setup:
            return

        cm.clear_brillouin_fits(calib_key)
        for frame_num, spectrum in enumerate(spectra):
            for region_key, region in enumerate(regions):
                xdata = np.arange(len(spectrum))
                w0s, fwhms, intensities, offset = \
                    fit_lorentz_region(
                        region,
                        xdata,
                        spectrum,
                        setup.calibration.num_brillouin_samples
                    )
                cm.add_brillouin_fit(calib_key, region_key, frame_num,
                                     w0s, fwhms, intensities, offset)

    def expected_frequencies(self, calib_key=None, current_frame=None):
        cm = self.session.calibration_model()

        if calib_key not in cm.vipa_params or \
                current_frame > len(cm.vipa_params[calib_key]) - 1:
            return None

        return self.session.setup.calibration.shifts \
            + self.session.setup.calibration.orders \
            * cm.vipa_params[calib_key][current_frame][3]


class PeakSelectionController(object):

    def __init__(self):
        self.session = Session.get_instance()
        return

    def add_brillouin_region_frequency(self, region_frequency):
        cm = self.session.calibration_model()
        # We use the first measurement image here
        time = self.session.get_payload_time('0')
        region_pix = tuple(
            cm.get_position_by_time(time, list(region_frequency)))

        psm = self.session.peak_selection_model()
        psm.add_brillouin_region(region_pix)

    def add_rayleigh_region_frequency(self, region_frequency):
        cm = self.session.calibration_model()
        # We use the first measurement image here
        time = self.session.get_payload_time('0')
        region_pix = tuple(
            cm.get_position_by_time(time, list(region_frequency)))

        psm = self.session.peak_selection_model()
        psm.add_rayleigh_region(region_pix)


class EvaluationController(ImageController):

    def __init__(self, *args, **kwargs):
        session = Session.get_instance()
        super(EvaluationController, self).__init__(
            model=session.evaluation_model,
            get_image=session.get_payload_image,
            get_time=session.get_payload_time,
            get_exposure=session.get_payload_exposure
        )
        return

    def evaluate(self, abort=None, count=None, max_count=None):
        em = self.session.extraction_model()
        if not em:
            if max_count is not None:
                max_count.value = -1
            return

        cm = self.session.calibration_model()
        if not cm:
            if max_count is not None:
                max_count.value = -1
            return

        pm = self.session.peak_selection_model()
        if not pm:
            if max_count is not None:
                max_count.value = -1
            return

        evm = self.session.evaluation_model()
        if not evm:
            if max_count is not None:
                max_count.value = -1
            return

        image_keys = self.session.get_image_keys(True)

        if max_count is not None:
            max_count.value = len(image_keys)

        brillouin_regions = pm.get_brillouin_regions()
        rayleigh_regions = pm.get_rayleigh_regions()

        resolution = self.session.get_payload_resolution()

        # Get first spectrum to find number of images
        spectra, _, _ = self.extract_spectra('0')

        if not spectra:
            if max_count is not None:
                max_count.value = -1
            return

        # We create a variable for this value here,
        # so changing nr_brillouin_peaks during evaluation
        # does not create issues
        nr_brillouin_peaks = evm.nr_brillouin_peaks
        evm.initialize_results_arrays({
            # measurement points in x direction
            'dim_x': resolution[0],
            # measurement points in y direction
            'dim_y': resolution[1],
            # measurement points in z direction
            'dim_z': resolution[2],
            # number of images per measurement point
            'nr_images': len(spectra),
            # number of Brillouin regions
            'nr_brillouin_regions': len(brillouin_regions),
            # number of peaks to fit per region
            'nr_brillouin_peaks': nr_brillouin_peaks,
            # number of Rayleigh regions
            'nr_rayleigh_regions': len(rayleigh_regions),
        })

        pool_size = mp.cpu_count() * 2
        pool = mp.Pool(processes=pool_size)
        # Initialize the Rayleigh shift
        # used for compensating drifts
        rayleigh_peak_initial =\
            np.nan * np.ones((len(rayleigh_regions)))
        rayleigh_shift = 0
        # Loop over all measurement positions
        for idx, image_key in enumerate(image_keys):
            # Calculate the indices for the given key
            (ind_x, ind_y, ind_z) =\
                self.get_indices_from_key(resolution, image_key)

            if (abort is not None) and abort.value:
                calculate_derived_values()
                if max_count is not None:
                    max_count.value = -1
                return
            spectra, times, intensities =\
                self.extract_spectra(image_key)
            if spectra is None:
                continue
            evm.results['time'][ind_x, ind_y, ind_z, :, 0, 0] =\
                times
            evm.results['intensity'][ind_x, ind_y, ind_z, :, 0, 0] =\
                intensities

            # We shift the evaluated regions here
            # to compensate for eventual drift
            brillouin_regions_shifted = [
                tuple([val + rayleigh_shift for val in region])
                for region in brillouin_regions]
            rayleigh_regions_shifted = [
                tuple([val + rayleigh_shift for val in region])
                for region in rayleigh_regions]

            # Pack the data for parallel processing
            regions = brillouin_regions_shifted + rayleigh_regions_shifted
            packed_data = zip(irepeat(spectra), regions)
            # Process it
            results = pool.starmap(self.fit_spectra, packed_data)

            # Unpack the results
            for frame_num, spectrum in enumerate(spectra):
                for region_key, _ in enumerate(brillouin_regions):
                    ind = (ind_x, ind_y, ind_z,
                           frame_num, region_key, 0)
                    evm.results['brillouin_peak_position'][ind] =\
                        results[region_key][frame_num][0]
                    evm.results['brillouin_peak_fwhm'][ind] =\
                        results[region_key][frame_num][1]
                    evm.results['brillouin_peak_intensity'][ind] =\
                        results[region_key][frame_num][2]

                for region_key, _ \
                        in enumerate(rayleigh_regions,
                                     start=len(brillouin_regions)):
                    ind = (ind_x, ind_y, ind_z, frame_num,
                           region_key - len(brillouin_regions))
                    evm.results['rayleigh_peak_position'][ind] =\
                        results[region_key][frame_num][0]
                    evm.results['rayleigh_peak_fwhm'][ind] =\
                        results[region_key][frame_num][1]
                    evm.results['rayleigh_peak_intensity'][ind] =\
                        results[region_key][frame_num][2]

            # We can only do a multi-peak fit after the single-peak
            # Rayleigh fit is done, because we have to know the
            # Rayleigh peak positions in GHz in order to convert
            # the multi-peak fit bounds given in GHz into the position
            # in pixels.
            if nr_brillouin_peaks > 1:
                ind =\
                    (ind_x, ind_y, ind_z, slice(None), slice(None), 0)
                rayleigh_peaks = np.transpose(
                    evm.results['rayleigh_peak_position'][ind]
                )
                bounds = self.create_bounds(
                    brillouin_regions_shifted,
                    times,
                    rayleigh_peaks
                )
                if bounds is not None:
                    packed_data_multi_peak =\
                        zip(irepeat(spectra),
                            brillouin_regions_shifted,
                            irepeat(nr_brillouin_peaks),
                            bounds)
                else:
                    packed_data_multi_peak =\
                        zip(irepeat(spectra),
                            brillouin_regions_shifted,
                            irepeat(nr_brillouin_peaks),
                            irepeat(bounds))
                # Process it
                results_multi_peak = pool.starmap(
                    self.fit_spectra, packed_data_multi_peak)

                for frame_num, spectrum in enumerate(spectra):
                    for region_key, _ in enumerate(
                            brillouin_regions):
                        ind = (ind_x, ind_y, ind_z,
                               frame_num, region_key,
                               slice(1, nr_brillouin_peaks+1))
                        evm.results[
                            'brillouin_peak_position'][ind] = \
                            results_multi_peak[
                                region_key][frame_num][0]
                        evm.results[
                            'brillouin_peak_fwhm'][ind] = \
                            results_multi_peak[
                                region_key][frame_num][1]
                        evm.results[
                            'brillouin_peak_intensity'][ind] = \
                            results_multi_peak[
                                region_key][frame_num][2]

            # Calculate the shift of the Rayleigh peaks,
            # in order to follow the peaks in case of a drift
            rayleigh_peak_current = np.nanmean(
                evm.results['rayleigh_peak_position'][
                    ind_x, ind_y, ind_z, :, :, 0], axis=0)
            # If we haven't found a valid Rayleigh peak position,
            # but the current one is valid, use it
            if not np.isnan(rayleigh_peak_current).all()\
                    and np.isnan(rayleigh_peak_initial).all():
                rayleigh_peak_initial = rayleigh_peak_current
            shift = np.nanmean(rayleigh_peak_current - rayleigh_peak_initial)
            if not np.isnan(shift):
                rayleigh_shift = round(shift)

            if count is not None:
                count.value += 1

            # Calculate the derived values every ten steps
            if not (idx % 10):
                calculate_derived_values()

        pool.close()
        pool.join()
        calculate_derived_values()

        return

    @staticmethod
    def fit_spectra(spectra, region, nr_peaks=1, bounds_w0=None):
        fits = []
        for frame_num, spectrum in enumerate(spectra):
            xdata = np.arange(len(spectrum))
            if bounds_w0 is None:
                fit = fit_lorentz_region(region, xdata, spectrum,
                                         nr_peaks)
            else:
                fit = fit_lorentz_region(region, xdata, spectrum,
                                         nr_peaks, bounds_w0[frame_num])
            fits.append(fit)
        return fits

    def create_bounds(self, brillouin_regions, times, rayleigh_peaks):
        """
        This function converts the bounds settings into
        a bounds object for the fitting function
        Allowed parameters for the bounds settings are
        - 'min'/'max'   -> Will be converted to the respective
            lower or upper limit of the given region
        - '-Inf', 'Inf' -> Will be converted to -np.Inf or np.Inf
        - number [GHz]  -> Will be converted into the pixel
            position of the given frequency
        Parameters
        ----------
        brillouin_regions
        times
        rayleigh_peaks

        Returns
        -------

        """
        cm = self.session.calibration_model()
        evm = self.session.evaluation_model()
        bounds = evm.bounds
        if bounds is None:
            return None

        # We need a Rayleigh peak position for
        # every combination of brillouin_region and time
        if rayleigh_peaks.shape != (len(brillouin_regions), len(times)):
            return None

        w0_bounds = []
        # We have to create a separate bound for every region
        for region_idx, region in enumerate(brillouin_regions):
            local_time = []
            for time_idx, time in enumerate(times):
                # In case this is an Anti-Stokes peak, we find the peak
                # with the higher frequency on the left hand side and
                # have to flip the bounds
                is_anti_stokes = np.mean(region) <\
                                 rayleigh_peaks[region_idx][time_idx]
                f_rayleigh = cm.get_frequency_by_time(
                    time,
                    rayleigh_peaks[region_idx][time_idx]
                )[()]

                local_bound = []
                for bound in bounds:
                    local_limit = []
                    for limit in bound:
                        if limit.lower() == 'min':
                            val = region[is_anti_stokes]
                        elif limit.lower() == 'max':
                            val = region[not is_anti_stokes]
                        elif limit.lower() == '-inf':
                            val = -((-1) ** is_anti_stokes)\
                                          * np.Inf
                        elif limit.lower() == 'inf':
                            val = ((-1) ** is_anti_stokes)\
                                          * np.Inf
                        else:
                            # Try to convert the value in GHz into
                            # a value in pixel depending on the time
                            try:
                                f = ((-1) ** is_anti_stokes)\
                                          * 1e9 * float(limit) + f_rayleigh
                                val = cm.get_position_by_time(time, f)[()]
                            except BaseException:
                                val = np.Inf

                        local_limit.append(val)
                    # Check that the bounds are sorted ascendingly
                    # (for anti-stokes, they might not).
                    local_limit.sort()
                    local_bound.append(local_limit)
                local_time.append(local_bound)
            w0_bounds.append(local_time)

        return w0_bounds

    def get_data(self, parameter_key, brillouin_peak_index=0):
        """
        This function returns the evaluated data,
        its positions, dimensionality and labels
        Parameters
        ----------
        parameter_key: str
            The key of the parameter requested.
            See bmlab.model.evaluation_model.get_parameter_keys()
        brillouin_peak_index: int
            The index of the Brillouin peak to show in case
            we did a multi-peak fit
            0: the single-peak fit
            1:nr_brillouin_peaks: the multi-peak fits
            nr_brillouin_peaks+1: all multi-peak fits average
            nr_brillouin_peaks+1: all multi-peak fits weighted average

        Returns
        -------
        data: np.ndarray
            The data to show. This is always a 3-dimensional array.
        positions: list
            This is a list of length 3 containing ndarrays with
            the spatial positions of the data points.
        dimensionality: int
            Whether it's a 0, 1, 2, or 3D measurement
        labels: list
            The labels of the positions
        """
        resolution = self.session.get_payload_resolution()

        dimensionality = sum(np.array(resolution) > 1)

        # Get the positions and squeeze them
        pos = self.session.get_payload_positions()

        positions = list(pos.values())
        labels = list(map(lambda l: r'$' + l + '$ [$\\mu$m]', ['x', 'y', 'z']))

        evm = self.session.evaluation_model()
        data = evm.results[parameter_key]

        # Ensure that we always get the expected shape
        # (even if the array was not initialized yet)
        if data.size == 0:
            data = np.empty(resolution)
            data[:] = np.nan

        # Slice the appropriate Brillouin peak if necessary and possible
        if data.ndim >= 6:
            nr_peaks_stored = data.shape[5]
            if nr_peaks_stored > 1\
                    and brillouin_peak_index < nr_peaks_stored + 2:
                if brillouin_peak_index < nr_peaks_stored:
                    sliced = data[:, :, :, :, :, brillouin_peak_index]
                # Average all multi-peak fits
                if brillouin_peak_index == nr_peaks_stored:
                    sliced = data[:, :, :, :, :, 1:]
                # Weighted average of all multi-peak fits
                if brillouin_peak_index == nr_peaks_stored + 1:
                    weight =\
                        evm.results[
                            'brillouin_peak_intensity'][:, :, :, :, :, 1:]\
                        * evm.results['brillouin_peak_fwhm'][:, :, :, :, :, 1:]
                    sliced = \
                        np.nansum(data[:, :, :, :, :, 1:] * weight, axis=5)\
                        / np.nansum(weight, axis=5)
            else:
                sliced = data[:, :, :, :, :, 0]
        else:
            sliced = data

        # Average all non-spatial dimensions.
        # Do not show warning which occurs when a slice contains only NaNs.
        with warnings.catch_warnings():
            warnings.filterwarnings(
                action='ignore',
                message='Mean of empty slice'
            )
            data = np.nanmean(
                sliced,
                axis=tuple(range(3, sliced.ndim))
            )

        # Scale the date in case of GHz
        data = evm.parameters[parameter_key]['scaling'] * data

        return data, positions, dimensionality, labels

    @staticmethod
    def get_key_from_indices(resolution, ind_x, ind_y, ind_z):
        if len(resolution) != 3:
            raise ValueError('resolution has wrong dimension')
        if ind_x >= resolution[0]:
            raise IndexError('x index out of range')
        if ind_y >= resolution[1]:
            raise IndexError('y index out of range')
        if ind_z >= resolution[2]:
            raise IndexError('z index out of range')
        return str(ind_z * (resolution[0] * resolution[1])
                   + ind_y * resolution[0] + ind_x)

    @staticmethod
    def get_indices_from_key(resolution, key):
        key = int(key)
        ind_z = floor(key / (resolution[0] * resolution[1]))
        ind_y = floor(
            (key - ind_z * (resolution[0] * resolution[1])) / resolution[0])
        ind_x = (key % (resolution[0] * resolution[1])) % resolution[0]
        # ind_y = (key - ind_x) % resolution[0]
        if ind_x >= resolution[0]\
                or ind_y >= resolution[1]\
                or ind_z >= resolution[2]:
            raise ValueError('Invalid key')
        return ind_x, ind_y, ind_z


def calculate_derived_values():
    """
    We calculate the derived parameters here:
    - Brillouin shift [pix]
    - Brillouin shift [GHz]
    - Brillouin peak width [GHz]
    - Rayleigh peak width [GHz]
    """
    session = Session.get_instance()
    evm = session.evaluation_model()
    if not evm:
        return

    if len(evm.results['time']) == 0:
        return

    time = evm.results['time']
    shape_brillouin = evm.results['brillouin_peak_position'].shape
    shape_rayleigh = evm.results['rayleigh_peak_position'].shape
    # If we have the same number of Rayleigh and Brillouin regions,
    # we can simply subtract the two arrays (regions are always
    # sorted by center in the peak selection model, so corresponding
    # regions should be at the same array index)
    if shape_brillouin[4] == shape_rayleigh[4]:
        evm.results['brillouin_shift'] = abs(
            evm.results['brillouin_peak_position'] -
            evm.results['rayleigh_peak_position']
        )

        # Calculate shift in GHz
        shift = calculate_shift_f(
            time,
            evm.results['brillouin_peak_position'],
            evm.results['rayleigh_peak_position']
        )
        if shift is not None:
            evm.results['brillouin_shift_f'] = shift

    # Having a different number of Rayleigh and Brillouin regions
    # doesn't really make sense. But in case I am missing something
    # here, we assign each Brillouin region the nearest (by center)
    # Rayleigh region.
    else:
        psm = session.peak_selection_model()
        if not psm:
            return
        brillouin_centers = list(map(np.mean, psm.get_brillouin_regions()))
        rayleigh_centers = list(map(np.mean, psm.get_rayleigh_regions()))

        for idx in range(len(brillouin_centers)):
            # Find corresponding (nearest) Rayleigh region
            d = list(
                map(
                    lambda x: abs(x - brillouin_centers[idx]),
                    rayleigh_centers
                )
            )
            idx_r = d.index(min(d))
            evm.results['brillouin_shift'][:, :, :, :, idx, :] = abs(
                evm.results[
                    'brillouin_peak_position'][:, :, :, :, idx, :] -
                evm.results[
                    'rayleigh_peak_position'][:, :, :, :, idx_r, :]
            )

            # Calculate shift in GHz
            shift = calculate_shift_f(
                time[:, :, :, :, 0, :],
                evm.results['brillouin_peak_position'][:, :, :, :, idx, :],
                evm.results['rayleigh_peak_position'][:, :, :, :, idx_r, :]
            )
            if shift is not None:
                evm.results['brillouin_shift_f'][:, :, :, :, idx, :] =\
                    shift

    # Calculate FWHM in GHz
    fwhm_brillouin = calculate_fwhm_f(
        time,
        evm.results['brillouin_peak_position'],
        evm.results['brillouin_peak_fwhm']
    )
    if fwhm_brillouin is not None:
        evm.results['brillouin_peak_fwhm_f'] = fwhm_brillouin

    fwhm_rayleigh = calculate_fwhm_f(
        time,
        evm.results['rayleigh_peak_position'],
        evm.results['rayleigh_peak_fwhm']
    )
    if fwhm_brillouin is not None:
        evm.results['rayleigh_peak_fwhm_f'] = fwhm_rayleigh


def calculate_shift_f(time, brillouin_position, rayleigh_position):
    session = Session.get_instance()
    cm = session.calibration_model()
    if not cm:
        return np.full(np.shape(brillouin_position), np.nan)
    brillouin_peak_f = cm.get_frequency_by_time(
        time,
        brillouin_position
    )
    rayleigh_peak_f = cm.get_frequency_by_time(
        time,
        rayleigh_position
    )

    if (brillouin_peak_f is not None) and \
            (rayleigh_peak_f is not None):
        return abs(
            brillouin_peak_f -
            rayleigh_peak_f
        )
    return np.full(np.shape(brillouin_position), np.nan)


def calculate_fwhm_f(time, peak_position, peak_fwhm):
    session = Session.get_instance()
    cm = session.calibration_model()
    if not cm:
        return np.full(np.shape(peak_position), np.nan)
    brillouin_peak_right_slope_f = cm.get_frequency_by_time(
        time,
        peak_position + peak_fwhm/2
    )
    brillouin_peak_left_slope_f = cm.get_frequency_by_time(
        time,
        peak_position - peak_fwhm/2
    )

    if (brillouin_peak_right_slope_f is not None) and \
            (brillouin_peak_left_slope_f is not None):
        return abs(
            brillouin_peak_right_slope_f -
            brillouin_peak_left_slope_f
        )
    return np.full(np.shape(peak_position), np.nan)


class Controller(object):

    def __init__(self):
        self.session = Session.get_instance()
        return

    def evaluate(self, filepath, setup, orientation,
                 brillouin_regions, rayleigh_regions, repetitions=None):
        # Load data file
        self.session.set_file(filepath)

        # Evaluate all repetitions if not requested differently
        if repetitions is None:
            repetitions = self.session.file.repetition_keys()

        for repetition in repetitions:
            # Select repetition
            self.session.set_current_repetition(repetition)
            self.session.set_setup(setup)

            # Set orientation
            self.session.orientation = orientation

            ec = ExtractionController()
            cc = CalibrationController()
            psc = PeakSelectionController()
            evc = EvaluationController()

            # First add all extraction points because this
            # can influence the extraction for other calibrations
            ec.find_points_all()

            # Then do the calibration
            for calib_key in self.session.get_calib_keys():
                cc.find_peaks(calib_key)

                cc.calibrate(calib_key)

            for region in brillouin_regions:
                psc.add_brillouin_region_frequency(region)

            for region in rayleigh_regions:
                psc.add_rayleigh_region_frequency(region)

            evc.evaluate()

        return self.session


class ExportController(object):

    def __init__(self):
        return

    @staticmethod
    def get_configuration():
        return {
            'fluorescence': {
                'export': True,
            },
            'fluorescenceCombined': {
                'export': True,
            },
            'brillouin': {
                'export': True,
                'parameters': ['brillouin_shift_f'],
                'brillouin_shift_f': {
                    'cax': ('min', 'max'),
                }
            },
        }

    def export(self, configuration=None):
        if not configuration:
            configuration = self.get_configuration()

        FluorescenceExport().export(configuration)
        FluorescenceCombinedExport().export(configuration)

        # BrillouinExport needs the EvaluationController
        # to nicely get the data, so we provide it here.
        # Not really nice, but importing it in BrillouinExport
        # leads to a circular dependency.
        BrillouinExport(EvaluationController()).export(configuration)



import pathlib

import pytest

from bmlab.session import Session, get_valid_source,\
    get_source_file_path, get_session_file_path, BmlabInvalidFileError


def data_file_path(file_name):
    return pathlib.Path(__file__).parent / 'data' / file_name


def test_get_source_file_path():
    assert get_source_file_path(data_file_path('EvalData/Water.h5')) \
           == data_file_path('RawData/Water.h5')
    assert get_source_file_path(data_file_path('Water.session.h5'))\
           == data_file_path('Water.h5')


def test_get_session_file_path():
    assert get_session_file_path(data_file_path('RawData/Water.h5')) \
           == data_file_path('EvalData/Water.h5')
    assert get_session_file_path(data_file_path('Water.h5'))\
           == data_file_path('Water.session.h5')


def test_get_valid_source_file():
    # File from BrillouinAcquisition
    assert get_valid_source(data_file_path('Water.h5')) \
           == data_file_path('Water.h5')
    # Session file from bmlab<0.0.14
    assert get_valid_source(data_file_path('1D-x.session.h5'))\
           == data_file_path('1D-x.h5')
    # Session file from bmlab>=0.0.14
    assert get_valid_source(data_file_path('1D-y.session.h5'))\
           == data_file_path('1D-y.h5')

    # Non-existent source data file
    with pytest.raises(FileNotFoundError) as err:
        get_valid_source(data_file_path('Unavailable.h5'))
    assert err.typename == 'FileNotFoundError'
    assert 'Unavailable.h5' in str(err.value.filename)

    # Session file without source data file
    with pytest.raises(FileNotFoundError) as err:
        get_valid_source(data_file_path('lonely.session.h5'))
    assert err.typename == 'FileNotFoundError'
    assert 'lonely.h5' in str(err.value.filename)

    # Session file with invalid source data file
    with pytest.raises(BmlabInvalidFileError) as err:
        get_valid_source(data_file_path('invalid.session.h5'))
    assert err.typename == 'BmlabInvalidFileError'
    assert 'invalid.h5' in str(err.value.filename)

    # Source data file with invalid session file
    with pytest.raises(BmlabInvalidFileError) as err:
        get_valid_source(data_file_path('invalid2.session.h5'))
    assert err.typename == 'BmlabInvalidFileError'
    assert 'invalid2.session.h5' in str(err.value.filename)

    # Invalid source data file
    with pytest.raises(BmlabInvalidFileError) as err:
        get_valid_source(data_file_path('invalid.h5'))
    assert err.typename == 'BmlabInvalidFileError'
    assert 'invalid.h5' in str(err.value.filename)


def test_session_is_singleton():
    session = Session.get_instance()

    with pytest.raises(Exception):
        Session()

    assert session

    id_session = id(session)

    session = Session.get_instance()
    assert id(session) == id_session


def test_session_initializes():
    session = Session.get_instance()
    session.set_file(data_file_path('Water.h5'))

    assert len(session.extraction_models) == 1
    assert len(session.calibration_models) == 1
    assert '0' in session.calibration_models.keys()


def test_clear_session():
    # Arrange; set up session
    session = Session.get_instance()
    session.set_file(data_file_path('Water.h5'))
    session.orientation.rotation = 1
    session.set_current_repetition('0')

    # Act
    session.clear()

    # Assert
    assert session.file is None
    assert session.orientation.rotation == 0
    assert session.current_repetition() is None


def test_session_aborted_repetition():
    session = Session.get_instance()
    session.set_file(data_file_path('aborted_repetition0.h5'))
    session.set_current_repetition('0')
    session.set_current_repetition('1')


def test_session_get_calib_keys():
    session = Session.get_instance()
    session.set_file(data_file_path('2D-xy.h5'))
    session.set_current_repetition('0')

    keys = session.get_calib_keys()
    assert len(keys) == 1
    keys = session.get_calib_keys(True)
    assert keys == ['1']


def test_session_get_image_keys():
    session = Session.get_instance()
    session.set_file(data_file_path('2D-xy.h5'))
    session.set_current_repetition('0')

    image_keys = session.get_image_keys()
    assert len(image_keys) == 15
    image_keys = session.get_image_keys(True)
    assert image_keys == \
           ['0', '3', '6', '9', '12',
            '1', '4', '7', '10', '13',
            '2', '5', '8', '11', '14']



import pathlib

import numpy as np
import pytest

from bmlab.image import set_orientation, find_max_in_radius
from bmlab.image import extract_lines_along_arc
from bmlab.geometry import Circle


def data_file_path(file_name):
    return pathlib.Path(__file__).parent / 'data' / file_name


def test_set_orientation_valid_argument():
    test_image = np.zeros([2, 3, 3, 4])
    with pytest.raises(ValueError):
        set_orientation(test_image)


def test_set_orientation():
    # Test two dimensional case
    test_image = np.zeros([2, 3])
    test_image[0, 0] = 1

    test_image_a = set_orientation(test_image, 1)
    test_image_b = set_orientation(test_image, 0, True, True)

    np.testing.assert_array_equal(test_image_a, np.array([[0., 1.],
                                                          [0., 0.],
                                                          [0., 0.]]))

    np.testing.assert_array_equal(test_image_b, np.array([[0., 0., 0.],
                                                          [0., 0., 1.]]))

    # Test three dimensional case
    test_images = np.zeros([3, 2, 3])
    test_images[0, 0, 0] = 1
    test_images[1, 0, 2] = 4

    test_images_a = set_orientation(test_images, 1)
    actual_a = np.zeros([3, 3, 2])
    actual_a[0, :, :] = set_orientation(np.squeeze(test_images[0, :, :]), 1)
    actual_a[1, :, :] = set_orientation(np.squeeze(test_images[1, :, :]), 1)

    test_images_b = set_orientation(test_images, 0, True, True)
    actual_b = np.zeros([3, 2, 3])
    actual_b[0, :, :] = set_orientation(
        np.squeeze(test_images[0, :, :]), 0, True, True)
    actual_b[1, :, :] = set_orientation(
        np.squeeze(test_images[1, :, :]), 0, True, True)

    np.testing.assert_array_equal(test_images_a, actual_a)
    np.testing.assert_array_equal(test_images_b, actual_b)


def test_find_max_in_radius():
    img = np.zeros((100, 100))
    img[20, 30] = 1
    xy0 = 25, 35
    actual = find_max_in_radius(img, xy0, 15)
    expected = 20, 30
    assert actual == expected

    actual = find_max_in_radius(img, expected, 15)
    assert actual == expected


def test_extract_lines_along_arc():

    circle = Circle((0, 0), 100)
    x = y = np.arange(110, dtype=int)
    X, Y = np.meshgrid(x, y, indexing='ij')
    phis = np.linspace(0, np.pi/2., 30)
    img = np.sqrt(X**2 + Y**2)
    img_2 = X**2 + Y**2

    arc = []
    num_points = 3
    for phi in phis:
        mid_point, e_r = circle.point(phi)
        points = [mid_point + e_r *
                  k for k in np.arange(-num_points, num_points + 1)]
        arc.append(np.array(points))
    arc = np.array(arc)

    # Act
    actual = extract_lines_along_arc(img, arc)
    actual_2 = extract_lines_along_arc(img_2, arc)

    # Assert
    one = np.ones_like(phis)
    np.testing.assert_allclose(actual, 100*one, rtol=1.E-4)
    np.testing.assert_allclose(actual_2, 10**4 * one, rtol=1.E-3)

    # Quadratic scaling should shift the sum to the outside
    assert np.all(actual_2 > actual)



import logging

from bmlab.models.regions import regions_merge_add_region, regions_check_order
from bmlab.serializer import Serializer

logger = logging.getLogger(__name__)


class PeakSelectionModel(Serializer):

    def __init__(self):
        self.brillouin_regions = []
        self.rayleigh_regions = []

    def add_brillouin_region(self, region):
        region = tuple(round(x) for x in region)

        regions_merge_add_region(
            self.brillouin_regions, region)

        regions_check_order(self.brillouin_regions)

    def set_brillouin_region(self, index, region):
        region = tuple(round(x) for x in region)

        if index >= len(self.brillouin_regions):
            self.add_brillouin_region(region)
        else:
            self.brillouin_regions[index] = region

        regions_check_order(self.brillouin_regions)

    def get_brillouin_regions(self):
        return self.brillouin_regions

    def clear_brillouin_regions(self):
        self.brillouin_regions = []

    def add_rayleigh_region(self, region):
        region = tuple(round(x) for x in region)

        regions_merge_add_region(
            self.rayleigh_regions, region)

        regions_check_order(self.rayleigh_regions)

    def set_rayleigh_region(self, index, region):
        region = tuple(round(x) for x in region)

        if index >= len(self.rayleigh_regions):
            self.add_rayleigh_region(region)
        else:
            self.rayleigh_regions[index] = region

        regions_check_order(self.rayleigh_regions)

    def get_rayleigh_regions(self):
        return self.rayleigh_regions

    def clear_rayleigh_regions(self):
        self.rayleigh_regions = []



import logging
import numpy as np
from collections import OrderedDict

from bmlab.serializer import Serializer

logger = logging.getLogger(__name__)


class EvaluationModel(Serializer):

    def __init__(self):

        self.nr_brillouin_peaks = 1
        self.spectra = {}

        # @since 0.1.0
        self.parameters = self.get_default_parameters()
        # @since 0.1.8
        self.bounds = None

        self.results = {}
        for key in self.parameters.keys():
            self.results[key] = np.empty((0,))

    def post_deserialize(self):
        # Migrations from 0.0.13 to 0.1.0
        # Check that the parameters attribute is present
        # @since 0.1.0
        if not hasattr(self, 'parameters')\
                or not isinstance(self, OrderedDict):
            self.parameters = self.get_default_parameters()
        # Migrations from 0.1.7 to 0.1.8
        # Check that the bounds attribute is present
        # @since 0.1.8
        if not hasattr(self, 'bounds'):
            self.bounds = None

    @staticmethod
    def get_default_parameters():
        return OrderedDict({
            'brillouin_shift_f': {          # [GHz] Brillouin frequency shift
                'unit': 'GHz',
                'symbol': r'$\nu_\mathrm{B}$',
                'label': 'Brillouin frequency shift',
                'scaling': 1e-9,
            },
            'brillouin_shift': {            # [pix] Brillouin frequency shift
                'unit': 'pix',
                'symbol': r'$\nu_\mathrm{B}$',
                'label': 'Brillouin frequency shift',
                'scaling': 1,
            },
            'brillouin_peak_fwhm_f': {      # [GHz] Brillouin peak FWHM
                'unit': 'GHz',
                'symbol': r'$\Delta_\mathrm{B}$',
                'label': 'Brillouin peak width',
                'scaling': 1e-9,
            },
            'brillouin_peak_fwhm': {        # [pix] Brillouin peak FWHM
                'unit': 'pix',
                'symbol': r'$\Delta_\mathrm{B}$',
                'label': 'Brillouin peak width',
                'scaling': 1,
            },
            'brillouin_peak_position': {    # [pix] Brillouin peak position
                'unit': 'pix',
                'symbol': r'$s_\mathrm{B}$',
                'label': 'Brillouin peak position',
                'scaling': 1,
            },
            'brillouin_peak_intensity': {   # [a.u.] Brillouin peak intensity
                'unit': 'a.u.',
                'symbol': r'$I_\mathrm{B}$',
                'label': 'Brillouin peak intensity',
                'scaling': 1,
            },
            'rayleigh_peak_fwhm_f': {       # [GHz] Rayleigh peak FWHM
                'unit': 'GHz',
                'symbol': r'$\Delta_\mathrm{R}$',
                'label': 'Rayleigh peak width',
                'scaling': 1e-9,
            },
            'rayleigh_peak_fwhm': {         # [pix] Rayleigh peak FWHM
                'unit': 'pix',
                'symbol': r'$\Delta_\mathrm{R}$',
                'label': 'Rayleigh peak width',
                'scaling': 1,
            },
            'rayleigh_peak_position': {     # [pix] Rayleigh peak position
                'unit': 'pix',
                'symbol': r'$s_\mathrm{R}$',
                'label': 'Rayleigh peak position',
                'scaling': 1,
            },
            'rayleigh_peak_intensity': {    # [a.u.] Rayleigh peak intensity
                'unit': 'a.u.',
                'symbol': r'$I_\mathrm{R}$',
                'label': 'Rayleigh peak intensity',
                'scaling': 1,
            },
            'intensity': {                  # [a.u.] Overall intensity of image
                'unit': 'a.u.',
                'symbol': r'$I_\mathrm{total}$',
                'label': 'Intensity',
                'scaling': 1,
            },
            'time': {                       # [s] The time the measurement
                'unit': 's',                # point was taken at
                'symbol': r'$t$',
                'label': 'Time',
                'scaling': 1,
            },
        })

    def initialize_results_arrays(self, dims):
        shape_general = (
            dims['dim_x'],
            dims['dim_y'],
            dims['dim_z'],
            dims['nr_images'],
            1,  # We just add this so it matches the ndims of the
            1,  # Brillouin array and reshapes are reduced
        )

        self.results['intensity'] = np.empty(shape_general)
        self.results['intensity'][:] = np.nan

        self.results['time'] = np.empty(shape_general)
        self.results['time'][:] = np.nan

        # We always do a single-peak fit, plus a multi-peak fit if requested.
        # Hence, we have to store
        # (nr_brillouin_peaks + 1) peaks, if nr_brillouin_peaks > 1.
        nr_brillouin_peaks_to_store = dims['nr_brillouin_peaks']
        if dims['nr_brillouin_peaks'] > 1:
            nr_brillouin_peaks_to_store = nr_brillouin_peaks_to_store + 1

        shape_brillouin = (
            dims['dim_x'],
            dims['dim_y'],
            dims['dim_z'],
            dims['nr_images'],
            dims['nr_brillouin_regions'],
            nr_brillouin_peaks_to_store,
        )

        self.results['brillouin_peak_position'] = np.empty(shape_brillouin)
        self.results['brillouin_peak_position'][:] = np.nan

        self.results['brillouin_peak_fwhm'] = np.empty(shape_brillouin)
        self.results['brillouin_peak_fwhm'][:] = np.nan

        self.results['brillouin_peak_intensity'] = np.empty(shape_brillouin)
        self.results['brillouin_peak_intensity'][:] = np.nan

        self.results['brillouin_shift'] = np.empty(shape_brillouin)
        self.results['brillouin_shift'][:] = np.nan

        self.results['brillouin_shift_f'] = np.empty(shape_brillouin)
        self.results['brillouin_shift_f'][:] = np.nan

        self.results['brillouin_peak_fwhm_f'] = np.empty(shape_brillouin)
        self.results['brillouin_peak_fwhm_f'][:] = np.nan

        shape_rayleigh = (
            dims['dim_x'],
            dims['dim_y'],
            dims['dim_z'],
            dims['nr_images'],
            dims['nr_rayleigh_regions'],
            1,  # We just add this so it matches the ndims of the
                #  Brillouin array and reshapes are reduced
        )

        self.results['rayleigh_peak_position'] = np.empty(shape_rayleigh)
        self.results['rayleigh_peak_position'][:] = np.nan

        self.results['rayleigh_peak_fwhm'] = np.empty(shape_rayleigh)
        self.results['rayleigh_peak_fwhm'][:] = np.nan

        self.results['rayleigh_peak_intensity'] = np.empty(shape_rayleigh)
        self.results['rayleigh_peak_intensity'][:] = np.nan

        self.results['rayleigh_peak_fwhm_f'] = np.empty(shape_rayleigh)
        self.results['rayleigh_peak_fwhm_f'][:] = np.nan

    def set_spectra(self, image_key, spectra):
        self.spectra[image_key] = spectra

    def get_spectra(self, image_key):
        spectra = self.spectra.get(image_key)
        if spectra:
            return spectra
        return None

    def get_parameter_keys(self):
        return self.parameters

    def setNrBrillouinPeaks(self, nr_brillouin_peaks):
        self.nr_brillouin_peaks = nr_brillouin_peaks

        # Initialize the bounds if necessary
        if nr_brillouin_peaks > 1 and\
                (self.bounds is None or
                 len(self.bounds) is not nr_brillouin_peaks):
            self.bounds = [['min', 'max'] for _ in range(nr_brillouin_peaks)]



from bmlab.models.setup import Calibration
import numpy as np


def test_calibration_2_samples():
    shift_methanol = 3.6e9
    shift_water = 5.0e9

    calibration = Calibration(num_brillouin_samples=2,
                              shift_methanol=shift_methanol,
                              shift_water=shift_water)

    np.testing.assert_array_equal(
        calibration.shifts,
        [0,
         shift_methanol,
         shift_water,
         -1*shift_water,
         -1*shift_methanol,
         0]
    )
    np.testing.assert_array_equal(
        calibration.orders,
        [0, 0, 0, 1, 1, 1]
    )


def test_calibration_1_samples():
    shift_methanol = 3.6e9
    shift_water = 5.0e9

    calibration = Calibration(num_brillouin_samples=1,
                              shift_methanol=shift_methanol,
                              shift_water=shift_water)

    np.testing.assert_array_equal(
        calibration.shifts,
        [0,
         shift_methanol,
         -1*shift_methanol,
         0]
    )
    np.testing.assert_array_equal(
        calibration.orders,
        [0, 0, 1, 1]
    )


def test_calibration_set_shift():
    shift_methanol = 3.6e9
    shift_water = 5.0e9

    calibration = Calibration(num_brillouin_samples=2,
                              shift_methanol=shift_methanol,
                              shift_water=shift_water)

    shift_methanol_new = 3.0e9
    calibration.set_shift_methanol(shift_methanol_new)

    np.testing.assert_array_equal(
        calibration.shifts,
        [0,
         shift_methanol_new,
         shift_water,
         -1*shift_water,
         -1*shift_methanol_new,
         0]
    )



import numpy as np
from bmlab.geometry import Circle, Rectangle, discretize_arc


def test_circle():
    assert Circle((1, 1), 1).is_valid()
    assert not Circle((1, 1), 0).is_valid()
    assert not Circle((1, 1), np.nan).is_valid()
    assert not Circle((1, np.inf), 1).is_valid()
    assert not Circle((-np.inf, 0), 1).is_valid()


def test_intersect_circle_rectangle():
    circle = Circle((-1, -1), 2)
    rect = Rectangle((1, 1))
    actual = circle.intersection(rect)

    cut = 0.72955
    expected = [(0, cut), (cut, 0)]
    np.testing.assert_array_almost_equal(actual, expected, decimal=5)


def test_circle_angle():
    circle = Circle((0, 0), 1)

    actual = circle.angle((1, 0))
    np.testing.assert_almost_equal(actual, 0.0)

    actual = circle.angle((1, 1))
    np.testing.assert_almost_equal(actual, np.pi / 4.)

    actual = circle.angle((0, 1))
    np.testing.assert_almost_equal(actual, np.pi / 2.)

    actual = circle.angle((-1, 0))
    np.testing.assert_almost_equal(actual, np.pi)

    actual = circle.angle((0, -1))
    np.testing.assert_almost_equal(actual, 3 * np.pi / 2.)

    actual = circle.angle((1, -1))
    np.testing.assert_almost_equal(actual, 3 * np.pi / 2. + np.pi/4)

    circle = Circle((1, 0), 1)
    actual = circle.angle((1, 1))
    np.testing.assert_almost_equal(actual, np.pi / 2)


def test_circle_er():
    circle = Circle((0, 0), 1)

    np.testing.assert_allclose(circle.e_r(0), [1, 0])
    np.testing.assert_allclose(circle.e_r(np.pi / 2), [0, 1], atol=1e-12)


def test_circle_point():
    circle = Circle((0, 0), 1)
    actual, _ = circle.point(np.pi / 4.)
    sq2 = np.sqrt(2.)
    np.testing.assert_array_almost_equal(actual, (sq2/2., sq2/2.))

    actual = circle.point(0, True)
    np.testing.assert_array_equal(actual, (1, 0))


def test_discretize_arc():
    circle = Circle((0, 0), 50)
    img_shape = (100, 100)
    num_points = 100
    expected = np.linspace(np.pi/2., 0, num_points)
    actual = discretize_arc(circle, img_shape, num_points)
    np.testing.assert_array_equal(actual, expected)

    circle = Circle((0, 0), 70)
    img_shape = (100, 100)
    num_points = 2
    expected = np.linspace(np.pi/2., 0, num_points)
    actual = discretize_arc(circle, img_shape, num_points)
    np.testing.assert_array_equal(actual, expected)

    r = 70
    shift = (r**2 / 2)**0.5
    circle = Circle((shift, 0), r)
    img_shape = (200, 200)
    num_points = 2
    expected = np.linspace(np.pi / 2. + np.pi / 4., 0, num_points)
    actual = discretize_arc(circle, img_shape, num_points)
    np.testing.assert_array_equal(actual, expected)



import time
from functools import wraps
import logging

import numpy as np

logger = logging.getLogger(__name__)


def debug_timer(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        logger.debug('%s needed: %f s' % (func.__name__, end - start))
        return result
    return wrapper


def array_dump(name, arr):
    """
    A utility function for dumping numpy arrays to disk.
    Used for manually creating test data in debugging mode.
    """
    with open(name + '.npy', 'wb') as f:
        np.save(f, arr)



# flake8: noqa: F401
from .extraction_model import ExtractionModel
from .calibration_model import CalibrationModel
from .peak_selection_model import PeakSelectionModel
from .evaluation_model import EvaluationModel
from .orientation import Orientation
from . import setup



import pathlib

from bmlab.session import Session
from bmlab.controllers import ExportController

"""
Place this file in the same directory as your RawData/EvalData folders
and adjust the filename.
"""
filename = 'Brillouin.h5'

# Construct path to file
file_path = pathlib.Path(__file__).parent / 'RawData' / filename

session = Session.get_instance()

# Load the file to be used
session.set_file(file_path)

ec = ExportController()
# Get the default configuration to adjust
export_config = ec.get_configuration()
# Disable exporting fluorescence data
export_config['fluorescence']['export'] = False
# Set caxis of Brillouin data
export_config['brillouin']['brillouin_shift_f']['cax'] = (5.0, 5.8)
# Export the data
ec.export(export_config)



from pathlib import Path

import numpy as np

from bmlab.session import Session
from bmlab.controllers import ExtractionController
from bmlab.models import ExtractionModel, Orientation
from bmlab.models.setup import AVAILABLE_SETUPS


def test_add_point(mocker):
    mocker.patch('bmlab.session.Session.get_calibration_time',
                 return_value=0)

    em = ExtractionModel()
    mocker.patch('bmlab.session.Session.extraction_model', return_value=em)

    ec = ExtractionController()
    ec.add_point('0', (15, 15))

    points = em.get_points('0')

    assert len(points) == 1
    assert points[0] == (15, 15)

    ec.add_point('0', (20, 20))

    assert len(points) == 2
    assert points[0] == (15, 15)
    assert points[1] == (20, 20)


def test_set_point(mocker):
    mocker.patch('bmlab.session.Session.get_calibration_time',
                 return_value=0)

    em = ExtractionModel()
    mocker.patch('bmlab.session.Session.extraction_model', return_value=em)

    ec = ExtractionController()
    ec.add_point('0', (15, 15))

    points = em.get_points('0')

    assert len(points) == 1
    assert points[0] == (15, 15)

    ec.add_point('0', (20, 20))

    assert len(points) == 2
    assert points[0] == (15, 15)
    assert points[1] == (20, 20)

    ec.set_point('0', 0, (10, 10))

    assert len(points) == 2
    assert points[0] == (10, 10)
    assert points[1] == (20, 20)

    ec.set_point('0', 2, (30, 20))

    assert len(points) == 3
    assert points[0] == (10, 10)
    assert points[1] == (20, 20)
    assert points[2] == (30, 20)


def test_optimize_points(mocker):

    imgs = np.zeros((1, 100, 100), dtype=int)
    imgs[0, 19:22, 19:22] = 1
    imgs[0, 79:82, 79:82] = 1

    em = ExtractionModel()

    mocker.patch('bmlab.session.Session.extraction_model', return_value=em)
    mocker.patch('bmlab.session.Session.get_calibration_image',
                 return_value=imgs)
    mocker.patch('bmlab.session.Session.get_calibration_time',
                 return_value=0)

    ec = ExtractionController()
    ec.add_point('0', (15, 15))
    ec.add_point('0', (5, 85))

    opt_points = em.get_points('0')

    assert (15, 15) == opt_points[0]
    assert (5, 85) == opt_points[1]

    ec.set_point('0', 1, (75, 85))

    ec.optimize_points('0', radius=10)
    opt_points = em.get_points('0')

    assert (19, 20) == opt_points[0]
    assert (79, 80) == opt_points[1]


def test_distance_point_to_line():
    ec = ExtractionController()

    np.testing.assert_almost_equal(0, ec.distance_point_to_line(
        (0.5, 0.5), (1, 1), (0, 0)
    ))
    np.testing.assert_almost_equal(0.5, ec.distance_point_to_line(
        (0.5, 0.5), (0, 0), (0, 1)
    ))
    np.testing.assert_almost_equal(np.sqrt(0.5), ec.distance_point_to_line(
        (0, 1), (1, 1), (0, 0)
    ))


def test_find_points(mocker):

    imgs = np.ones((1, 400, 200), dtype=int)
    imgs[0, 19:22, 19:22] = 5
    imgs[0, 379:382, 19:22] = 5
    imgs[0, 19:22, 179:182] = 5
    imgs[0, 379:382, 179:182] = 5

    em = ExtractionModel()

    mocker.patch('bmlab.session.Session.extraction_model', return_value=em)
    mocker.patch('bmlab.session.Session.get_calibration_image',
                 return_value=imgs)
    mocker.patch('bmlab.session.Session.get_calibration_time',
                 return_value=0)

    ec = ExtractionController()

    ec.find_points('0', 1, 4, 10)
    opt_points = em.get_points('0')

    assert len(opt_points) == 2
    assert (19, 180) == opt_points[0]
    assert (379, 20) == opt_points[1]


def test_find_points_real_data():
    # Start session
    session = Session.get_instance()

    # Load data file
    session.set_file(Path(__file__).parent / 'data' / 'Water.h5')

    # Select repetition
    session.set_current_repetition('0')
    session.set_setup(AVAILABLE_SETUPS[0])

    # Set orientation
    session.orientation = Orientation(rotation=1, reflection={
        'vertically': False, 'horizontally': False
    })

    em = session.extraction_model()

    ec = ExtractionController()

    points = {
        '1': [
            (107, 293),
            (165, 237),
            (182, 218),
            (240, 154),
            (254, 137),
            (291, 92),
        ],
        '2': [
            (107, 293),
            (165, 237),
            (182, 218),
            (241, 153),
            (255, 137),
            (291, 92),
        ]
    }

    for calib_key in session.get_calib_keys():
        ec.find_points(calib_key)

        p = em.get_points(calib_key)

        assert p == points[calib_key]



import numpy as np
from scipy import interpolate

from bmlab.fits import fit_circle
from bmlab.geometry import Circle, discretize_arc
from bmlab.serializer import Serializer


class ExtractionModel(Serializer):

    def __init__(self):
        self.image_shape = None
        self.arc_width = 2  # [pix] the width of the extraction arc
        self.points = {}
        self.calib_times = {}
        self.positions = {}
        self.positions_interpolation = None

    def add_point(self, calib_key, time, xdata, ydata):
        if calib_key not in self.points:
            self.points[calib_key] = []
        self.points[calib_key].append((xdata, ydata))
        self.calib_times[calib_key] = time
        self.update_positions(calib_key)

    def set_point(self, calib_key, index, time, xdata, ydata):
        if calib_key not in self.points:
            self.points[calib_key] = []
        if index < len(self.points[calib_key]):
            self.points[calib_key][index] = (xdata, ydata)
        else:
            self.points[calib_key].append((xdata, ydata))
        self.calib_times[calib_key] = time
        self.update_positions(calib_key)

    def set_points(self, calib_key, time, points):
        self.calib_times[calib_key] = time
        self.points[calib_key] = points
        self.update_positions(calib_key)

    def get_points(self, calib_key):
        if calib_key in self.points:
            return self.points[calib_key]
        return []

    def get_time(self, calib_key):
        if calib_key in self.calib_times:
            return self.calib_times[calib_key]
        return []

    def clear_points(self, calib_key):
        self.points.pop(calib_key, None)
        self.calib_times.pop(calib_key, None)
        self.update_positions(calib_key)

    def post_deserialize(self):
        # Migrations from 0.1.10 to 0.2.0
        # Check that correct attributes are present
        # @since 0.2.0
        attributes_to_remove = [
            'circle_fits',
            'circle_fits_index',
            'circle_fits_interpolation',
            'extraction_angles',
            'extraction_angles_index',
            'extraction_angles_interpolation'
        ]
        for attribute in attributes_to_remove:
            if hasattr(self, attribute):
                delattr(self, attribute)
        if not hasattr(self, 'positions'):
            self.positions = {}
        if not hasattr(self, 'positions_interpolation'):
            self.positions_interpolation = None
        self.update_positions()

    def update_positions(self, key=None):
        if not hasattr(self, 'image_shape') or self.image_shape is None:
            return
        if not key:
            points = self.points.items()
        elif key in self.points:
            points = [(key, self.points[key])]
        else:
            points = []

        for calib_key, points in points:
            if len(points) >= 3:
                center, radius = fit_circle(points)
                # Check that we got a valid circle before continuing
                circle = Circle(center, radius)
                if not circle.valid:
                    continue

                phis = discretize_arc(circle, self.image_shape, num_points=500)
                if phis is None:
                    continue
                arc = self.get_arc_from_circle_phis(
                    circle, phis, self.arc_width)
                self.positions[calib_key] = arc
            # If we don't have enough points but positions
            # already present for this key, we have probably removed points
            # and clear the positions then
            elif calib_key in self.positions:
                self.positions.pop(calib_key, None)

        self.refresh_positions_interpolation()

    def refresh_positions_interpolation(self):
        # If there are no entries, we reset the interpolation
        if not self.calib_times:
            self.positions_interpolation = None
            return

        # Sort calibration keys by time
        sorted_keys = sorted(self.calib_times,
                             key=self.calib_times.get)

        # Create arrays to interpolate
        calib_times_array = []
        positions = []
        for calib_key in sorted_keys:
            # We might not have positions for all present keys yet
            if calib_key in self.positions:
                calib_times_array.append(self.calib_times[calib_key])
                positions.append(self.positions[calib_key])
        positions_array = np.array(positions)
        calib_times_array = np.array(calib_times_array)

        # If we only have one entry we cannot interpolate
        # and always return the same value
        if len(calib_times_array) < 1:
            self.positions_interpolation = None
        elif len(calib_times_array) == 1:
            self.positions_interpolation =\
                lambda time: positions_array[0]
        else:
            self.positions_interpolation =\
                interpolate.interp1d(
                    calib_times_array,
                    positions_array,
                    axis=0,
                    bounds_error=False,
                    fill_value=(positions_array[0], positions_array[-1])
                )

    def get_arc_by_calib_key(self, calib_key):
        """
        Returns the arc at which to interpolate the 2D image for
        the 1D spectrum

        Parameters
        ----------
        calib_key: the calibration number

        Returns
        -------
        The arc with pixel positions
        """
        arc = np.empty(0)
        try:
            if calib_key in self.positions:
                arc = self.positions.get(calib_key)
        finally:
            return arc

    def get_arc_by_time(self, time):
        """
        Returns the arc at which to interpolate the 2D image for
        the 1D spectrum

        Parameters
        ----------
        time: time point

        Returns
        -------
        The arc with pixel positions
        """
        arc = np.empty(0)
        try:
            arc = self.positions_interpolation(time)
        finally:
            return arc

    @staticmethod
    def get_arc_from_circle_phis(circle, phis, arc_width):
        pos_e_r = np.arange(-arc_width, arc_width + 1)[..., None]
        arc = np.ndarray((len(phis), len(pos_e_r), 2))
        for i, phi in enumerate(phis):
            mid_point, e_r = circle.point(phi)
            arc[i, :, :] = mid_point + e_r * pos_e_r
        return arc

    # TODO: This needs to be called automatically
    #  on file load or when the image orientation is changed
    def set_image_shape(self, shape):
        if not hasattr(self, 'image_shape') or self.image_shape != shape:
            self.image_shape = shape
            self.update_positions()

    def set_arc_width(self, width):
        self.arc_width = width
        self.update_positions()


class CircleFit(Serializer):

    def __init__(self, center, radius):
        self.center = center
        self.radius = radius


class ExtractionException(Exception):
    pass



import pathlib

import numpy as np

from bmlab.controllers import Controller
from bmlab.models import Orientation
from bmlab.models.setup import AVAILABLE_SETUPS


def run_pipeline():

    filepath = pathlib.Path(__file__).parent / 'data' / 'Water.h5'
    setup = AVAILABLE_SETUPS[0]
    orientation = Orientation(rotation=1, reflection={
            'vertically': False, 'horizontally': False
        })

    # Frequency ranges to evaluate in GHz
    brillouin_regions = [(4.0e9, 6.0e9), (9.0e9, 11.0e9)]
    rayleigh_regions = [(-2.0e9, 2.0e9), (13.0e9, 17.0e9)]

    session = Controller().evaluate(
        filepath,
        setup,
        orientation,
        brillouin_regions,
        rayleigh_regions
    )

    return session


def test_run_pipeline():
    session = run_pipeline()
    evm = session.evaluation_model()
    shift = evm.results['brillouin_shift_f']
    assert shift.size != 0
    np.testing.assert_allclose(shift, 5.03e9, atol=50E6)



"""
Module to perform common image operations.
"""

import logging

import numpy as np
from scipy.interpolate import interpolate
import warnings


class AutofindException(Exception):
    pass


logger = logging.getLogger(__name__)


def set_orientation(image, rotate=0, flip_ud=False, flip_lr=False):
    """
    Change the orientation of an image.

    The orientation of the raw images is setup dependent.

    Parameters
    ----------
    image: array_like
        Array of two or three dimensions (containing multiple
        images along the first dimension in case of 3D)
    rotate: int
        Number of times the array is rotated by 90 degrees (clockwise).
    flip_ud: bool
        If True, flip the image up-down
    flip_lr: bool
        If True, flip the image left-right

    Returns
    -------
    image: array_like
        The input image in the requested orientation. The order
        of the operations is equal to the order
        of the input arguments: rotate, flip up-down, flip left right

    See Also
    --------
    numpy.rot90, numpy.flipud, numpy.fliplr
    """
    if np.shape(np.shape(image)) == (2,):
        if rotate != 0:
            image = np.rot90(image, k=rotate, axes=(1, 0))

        if flip_ud:
            image = np.flipud(image)

        if flip_lr:
            image = np.fliplr(image)
    elif np.shape(np.shape(image)) == (3,):
        if rotate != 0:
            image = np.rot90(image, k=rotate, axes=(2, 1))

        if flip_ud:
            image = image[:, ::-1, :]

        if flip_lr:
            image = image[:, :, ::-1]
    else:
        raise ValueError(
            'Given Argument is not a two or three dimensional array'
        )

    return image


def find_max_in_radius(img, xy0, radius):
    """
    Returns the index of the max. value in a circle around given point.

    Parameters
    ----------
    img: numpy.ndarray (2D)
        the image data
    xy0: tuple
        x-y indices of point around which to search for maximum
    radius: float
        the radius of the search

    Returns
    -------
    out: tuple
        x-y indices of the point of max. value
    """
    x, y = list(range(img.shape[0])), list(range(img.shape[1]))
    X, Y = np.meshgrid(x, y, indexing='ij')
    x0, y0 = xy0
    flat_img = np.nan * np.ones_like(img)
    mask = (X - x0) ** 2 + (Y - y0) ** 2 <= radius ** 2
    flat_img[mask] = img[mask]
    peak_idx = np.nanargmax(flat_img)
    peak_x, peak_y = np.unravel_index(peak_idx, img.shape, order='C')
    return peak_x, peak_y


def extract_lines_along_arc(img, arc):
    if arc.ndim != 3 or arc.shape[2] != 2:
        return

    m, n = img.shape
    func = interpolate.RegularGridInterpolator(
        (np.arange(m), np.arange(n)), img, method='linear', bounds_error=False)

    # In case the arc crosses the edge of the image, the interpolated array
    # will contain rows with only NaNs leading to an empty slice warning
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action='ignore',
            message='Mean of empty slice'
        )
        return np.nanmean(func((arc[:, :, 0], arc[:, :, 1])), 1)



# flake8: noqa: F401
from . import file
from . import fits
from . import geometry
from . import image
from . import utils
from ._version import version as __version__
from .models import calibration_model
from .models import extraction_model
from .models import orientation
from .models import setup

from .session import Session


from bmlab.models.peak_selection_model import PeakSelectionModel


def test_peak_selection_model_add_brillouin_region():
    pm = PeakSelectionModel()
    pm.add_brillouin_region([4.9, 7])
    pm.add_brillouin_region((6, 9))

    brillouin_region_0 = pm.get_brillouin_regions()

    assert brillouin_region_0 == [(5, 9)]

    pm.add_brillouin_region((1, 3))

    assert brillouin_region_0 == [(1, 3), (5, 9)]

    pm.add_brillouin_region((2, 3))

    assert brillouin_region_0 == [(1, 3), (5, 9)]


def test_peak_selection_model_set_brillouin_region():
    pm = PeakSelectionModel()
    pm.add_brillouin_region((6, 9))
    pm.set_brillouin_region(0, [0.9, 3])
    pm.set_brillouin_region(1, [4, 5])

    brillouin_region_0 = pm.get_brillouin_regions()

    assert brillouin_region_0 == [(1, 3), (4, 5)]

    pm.set_brillouin_region(0, [7, 9])

    assert brillouin_region_0 == [(4, 5), (7, 9)]


def test_peak_selection_model_add_rayleigh_region():
    pm = PeakSelectionModel()
    pm.add_rayleigh_region([4.9, 7])
    pm.add_rayleigh_region((6, 9))

    rayleigh_region_0 = pm.get_rayleigh_regions()

    assert rayleigh_region_0 == [(5, 9)]

    pm.add_rayleigh_region((1, 3))

    assert rayleigh_region_0 == [(1, 3), (5, 9)]

    pm.add_rayleigh_region((2, 3))

    assert rayleigh_region_0 == [(1, 3), (5, 9)]


def test_peak_selection_model_set_rayleigh_region():
    pm = PeakSelectionModel()
    pm.add_rayleigh_region((6, 9))
    pm.set_rayleigh_region(0, [0.9, 3])
    pm.set_rayleigh_region(1, [4, 5])

    rayleigh_region_0 = pm.get_rayleigh_regions()

    assert rayleigh_region_0 == [(1, 3), (4, 5)]

    pm.set_rayleigh_region(0, [7, 9])

    assert rayleigh_region_0 == [(4, 5), (7, 9)]



import numpy as np

import bmlab.constants as constants
from bmlab.serializer import Serializer


class Setup(Serializer):

    def __init__(self, key, name, pixel_size, focal_length,
                 vipa, calibration):
        """

        Parameters
        ----------
        key: str
            ID for the setup
        name: str
            Name of setup
        pixel_size: float
            pixel size of the camera [m]
        lambda0: float
            laser wavelength [m]
        focal_length: float
            focal length of the lens behind the VIPA [m]
        vipa
        calibration
        """
        self.key = key
        self.name = name
        self.pixel_size = pixel_size
        self.lambda0 = vipa.lambda0
        self.f0 = constants.c / self.lambda0
        self.focal_length = focal_length
        self.vipa = vipa
        self.calibration = calibration
        self.VIPA_PARAMS = self.init_vipa_params()

    def init_vipa_params(self):
        p1 = (2 * np.pi * self.vipa.n * self.vipa.d *
              np.cos(self.vipa.theta)) / constants.c
        p2 = -(2 * np.pi * self.vipa.n * self.vipa.d *
               np.tan(self.vipa.theta)) /\
              (constants.c * self.focal_length) * np.sqrt(
            1 - (self.vipa.n * np.sin(self.vipa.theta)) ** 2)
        p3 = -np.pi / constants.c * self.vipa.d * np.cos(self.vipa.theta) /\
            (self.focal_length ** 2)

        return p1, p2, p3


class VIPA(Serializer):

    def __init__(self, d, n, theta, order, lambda0):
        """ Start values for VIPA fit

        Parameters
        ----------
        d : float
            width of the cavity [m]
        n : float
            refractive index of the cavity [-]
        theta : float
            angle [rad]
        order: int
            observed order of the VIPA spectrum
        """
        self.d = d
        self.n = n
        self.theta = theta
        self.order = order
        self.lambda0 = lambda0
        self.FSR = constants.c / (2 * self.n * self.d * np.cos(self.theta))
        self.m = round(constants.c/(self.lambda0 * self.FSR))


class Calibration(Serializer):

    def __init__(self, num_brillouin_samples, shift_methanol=None,
                 shift_water=None):
        """

        Parameters
        ----------
        num_brillouin_samples: int
            Number of samples
        shift_methanol: float
            ??
        shift_water: float
            ??
        """

        self.num_brillouin_samples = num_brillouin_samples
        self.shift_methanol = shift_methanol
        self.shift_water = shift_water
        self.shifts = np.array([])
        self.orders = np.array([])

        self.update_calibration()

    def set_shift_water(self, shift_water):
        self.shift_water = shift_water
        self.update_calibration()

    def set_shift_methanol(self, shift_methanol):
        self.shift_methanol = shift_methanol
        self.update_calibration()

    def update_calibration(self):
        # Construct array with the frequency shifts
        tmp = [self.shift_methanol, self.shift_water]
        self.shifts = np.full(2 + 2 * self.num_brillouin_samples, 0.0)
        for i in range(self.num_brillouin_samples):
            self.shifts[i + 1] = tmp[i]
            self.shifts[-1 * (i + 2)] = -1 * tmp[i]

        # The interference orders to which the peaks belong
        self.orders = np.full(2 + 2 * self.num_brillouin_samples, 0)
        self.orders[-(1 + self.num_brillouin_samples):] = 1


AVAILABLE_SETUPS = [
    Setup(key='S0',
          name='780 nm @ Biotec R340',
          pixel_size=6.5e-6,
          focal_length=0.2,
          vipa=VIPA(d=0.006743,
                    n=1.45367,
                    theta=0.8 * 2 * np.pi / 360,
                    order=0,
                    lambda0=780.24e-9),
          calibration=Calibration(num_brillouin_samples=2,
                                  shift_methanol=3.78e9,
                                  shift_water=5.066e9)),
    Setup(key='S1',
          name='780 nm @ Biotec R340 old',
          pixel_size=6.5e-6,
          focal_length=0.2,
          vipa=VIPA(d=0.006743,
                    n=1.45367,
                    theta=0.8 * 2 * np.pi / 360,
                    order=0,
                    lambda0=780.24e-9),
          calibration=Calibration(num_brillouin_samples=1,
                                  shift_methanol=3.78e9)),
    Setup(key='S2',
          name='532 nm @ Biotec R314',
          pixel_size=6.5e-6,
          focal_length=0.2,
          vipa=VIPA(d=0.003371,
                    n=1.46071,
                    theta=0.8 * 2 * np.pi / 360,
                    order=0,
                    lambda0=532e-9),
          calibration=Calibration(num_brillouin_samples=2,
                                  shift_methanol=5.54e9,
                                  shift_water=7.43e9))
]



import pathlib
import numpy as np

import pytest

from bmlab.controllers import EvaluationController, calculate_derived_values
from bmlab.models import CalibrationModel, EvaluationModel


def data_file_path(file_name):
    return pathlib.Path(__file__).parent / 'data' / file_name


def test_get_key_from_indices():
    resolution = (7, 8, 9)
    assert EvaluationController\
        .get_key_from_indices(resolution, 0, 0, 0) == '0'
    assert EvaluationController\
        .get_key_from_indices(resolution, 6, 0, 0) == '6'
    assert EvaluationController\
        .get_key_from_indices(resolution, 0, 1, 0) == '7'
    assert EvaluationController\
        .get_key_from_indices(resolution, 1, 1, 0) == '8'
    assert EvaluationController\
        .get_key_from_indices(resolution, 6, 7, 0) == '55'
    assert EvaluationController\
        .get_key_from_indices(resolution, 0, 0, 1) == '56'
    assert EvaluationController\
        .get_key_from_indices(resolution, 0, 1, 1) == '63'
    assert EvaluationController\
        .get_key_from_indices(resolution, 0, 0, 8) == '448'
    assert EvaluationController\
        .get_key_from_indices(resolution, 1, 0, 8) == '449'
    assert EvaluationController\
        .get_key_from_indices(resolution, 1, 1, 8) == '456'
    assert EvaluationController\
        .get_key_from_indices(resolution, 6, 7, 8) == '503'

    # Resolution is not three-dimensional
    with pytest.raises(ValueError) as err:
        EvaluationController \
            .get_key_from_indices((7, 8), 6, 7, 8)
    assert err.typename == 'ValueError'

    # x index is out or range
    with pytest.raises(IndexError) as err:
        EvaluationController \
            .get_key_from_indices((7, 8, 9), 7, 9, 9)
    assert err.typename == 'IndexError'


def test_get_indices_from_key():
    resolution = (7, 8, 9)
    assert EvaluationController\
        .get_indices_from_key(resolution, '0') == (0, 0, 0)
    assert EvaluationController\
        .get_indices_from_key(resolution, '6') == (6, 0, 0)
    assert EvaluationController\
        .get_indices_from_key(resolution, '7') == (0, 1, 0)
    assert EvaluationController\
        .get_indices_from_key(resolution, '8') == (1, 1, 0)
    assert EvaluationController\
        .get_indices_from_key(resolution, '55') == (6, 7, 0)
    assert EvaluationController\
        .get_indices_from_key(resolution, '56') == (0, 0, 1)
    assert EvaluationController\
        .get_indices_from_key(resolution, '63') == (0, 1, 1)
    assert EvaluationController\
        .get_indices_from_key(resolution, '448') == (0, 0, 8)
    assert EvaluationController\
        .get_indices_from_key(resolution, '449') == (1, 0, 8)
    assert EvaluationController\
        .get_indices_from_key(resolution, '456') == (1, 1, 8)
    assert EvaluationController\
        .get_indices_from_key(resolution, '503') == (6, 7, 8)

    # Key is invalid
    with pytest.raises(ValueError) as err:
        EvaluationController \
            .get_indices_from_key((7, 8, 9), '504')
    assert err.typename == 'ValueError'


def test_evaluate():
    evaluationcontroller = EvaluationController()
    evaluationcontroller.evaluate()


def test_calculate_derived_values_equal_region_count():
    evc = EvaluationController()
    evc.session.set_file(data_file_path('Water.h5'))
    evc.session.set_current_repetition('0')
    evm = evc.session.evaluation_model()

    """
    Test calculation in case of same number
    Brillouin and Rayleigh regions
    """
    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': 5,
        'dim_y': 5,
        'dim_z': 5,
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 2,
    })

    evm.results['brillouin_peak_position'][:] = 1
    evm.results['rayleigh_peak_position'][:] = 3

    calculate_derived_values()

    assert (evm.results['brillouin_shift'] == 2).all()

    evm.results['brillouin_peak_position'][:, :, :, :, 0, :] = 1
    evm.results['brillouin_peak_position'][:, :, :, :, 1, :] = 4
    evm.results['rayleigh_peak_position'][:, :, :, :, 0, :] = 3
    evm.results['rayleigh_peak_position'][:, :, :, :, 1, :] = 8

    calculate_derived_values()

    assert (evm.results['brillouin_shift'][:, :, :, :, 0, :] == 2).all()
    assert (evm.results['brillouin_shift'][:, :, :, :, 1, :] == 4).all()


def test_calculate_derived_values_equal_region_count_nr_peaks_2():
    evc = EvaluationController()
    evc.session.set_file(data_file_path('Water.h5'))
    evc.session.set_current_repetition('0')
    evm = evc.session.evaluation_model()

    """
    Test calculation in case of same number
    Brillouin and Rayleigh regions
    """
    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': 5,
        'dim_y': 5,
        'dim_z': 5,
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 2,
        'nr_rayleigh_regions': 2,
    })

    evm.results['brillouin_peak_position'][:] = 1
    evm.results['rayleigh_peak_position'][:] = 3

    calculate_derived_values()

    assert (evm.results['brillouin_shift'] == 2).all()

    evm.results['brillouin_peak_position'][:, :, :, :, 0, :] = 1
    evm.results['brillouin_peak_position'][:, :, :, :, 1, :] = 4
    evm.results['brillouin_peak_position'][:, :, :, :, 0, 1] = 2
    evm.results['brillouin_peak_position'][:, :, :, :, 1, 1] = 5
    evm.results['rayleigh_peak_position'][:, :, :, :, 0, :] = 3
    evm.results['rayleigh_peak_position'][:, :, :, :, 1, :] = 8

    calculate_derived_values()

    assert (evm.results['brillouin_shift'][:, :, :, :, 0, 0] == 2).all()
    assert (evm.results['brillouin_shift'][:, :, :, :, 1, 0] == 4).all()
    assert (evm.results['brillouin_shift'][:, :, :, :, 0, 1] == 1).all()
    assert (evm.results['brillouin_shift'][:, :, :, :, 1, 1] == 3).all()


def test_calculate_derived_values_different_region_count():
    evc = EvaluationController()
    evc.session.set_file(data_file_path('Water.h5'))
    evc.session.set_current_repetition('0')
    evm = evc.session.evaluation_model()

    """
    Test calculation in case of more Rayleigh than
    Brillouin regions
    """
    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': 5,
        'dim_y': 5,
        'dim_z': 5,
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 3,
    })

    evm.results['brillouin_peak_position'][:, :, :, :, 0, :] = 1
    evm.results['brillouin_peak_position'][:, :, :, :, 1, :] = 4
    evm.results['rayleigh_peak_position'][:, :, :, :, 0, :] = 5
    evm.results['rayleigh_peak_position'][:, :, :, :, 1, :] = 9
    evm.results['rayleigh_peak_position'][:, :, :, :, 2, :] = 10

    psm = evc.session.peak_selection_model()

    psm.add_brillouin_region((1, 2))
    psm.add_brillouin_region((4, 5))

    psm.add_rayleigh_region((0, 1))
    psm.add_rayleigh_region((6, 7))
    psm.add_rayleigh_region((8, 9))

    calculate_derived_values()

    assert (evm.results['brillouin_shift'][:, :, :, :, 0, :] == 4).all()
    assert (evm.results['brillouin_shift'][:, :, :, :, 1, :] == 5).all()


def test_calculate_derived_values_different_region_count_nr_peaks_2():
    evc = EvaluationController()
    evc.session.set_file(data_file_path('Water.h5'))
    evc.session.set_current_repetition('0')
    evm = evc.session.evaluation_model()

    """
    Test calculation in case of more Rayleigh than
    Brillouin regions
    """
    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': 5,
        'dim_y': 5,
        'dim_z': 5,
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 2,
        'nr_rayleigh_regions': 3,
    })

    evm.results['brillouin_peak_position'][:, :, :, :, 0, :] = 1
    evm.results['brillouin_peak_position'][:, :, :, :, 1, :] = 4
    evm.results['brillouin_peak_position'][:, :, :, :, 0, 1] = 2
    evm.results['brillouin_peak_position'][:, :, :, :, 1, 1] = 5
    evm.results['rayleigh_peak_position'][:, :, :, :, 0, :] = 5
    evm.results['rayleigh_peak_position'][:, :, :, :, 1, :] = 9
    evm.results['rayleigh_peak_position'][:, :, :, :, 2, :] = 10

    psm = evc.session.peak_selection_model()

    psm.add_brillouin_region((1, 2))
    psm.add_brillouin_region((4, 5))

    psm.add_rayleigh_region((0, 1))
    psm.add_rayleigh_region((6, 7))
    psm.add_rayleigh_region((8, 9))

    calculate_derived_values()

    assert (evm.results['brillouin_shift'][:, :, :, :, 0, 0] == 4).all()
    assert (evm.results['brillouin_shift'][:, :, :, :, 1, 0] == 5).all()
    assert (evm.results['brillouin_shift'][:, :, :, :, 0, 1] == 3).all()
    assert (evm.results['brillouin_shift'][:, :, :, :, 1, 1] == 4).all()


def test_get_data_0D(mocker):
    evc = EvaluationController()
    evc.session.set_file(data_file_path('0D.h5'))
    evc.session.set_current_repetition('0')

    evm = evc.session.evaluation_model()

    resolution = evc.session.get_payload_resolution()
    assert resolution == (1, 1, 1)

    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': resolution[0],
        'dim_y': resolution[1],
        'dim_z': resolution[2],
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 2,
    })
    evm.results['brillouin_shift_f'][:] = 1e9  # [Hz]

    data, positions, dimensionality, labels = evc.get_data('brillouin_shift_f')

    assert data == 1  # [GHz]
    assert positions[0][0, 0, 0] == -1
    assert positions[1][0, 0, 0] == -2
    assert positions[2][0, 0, 0] == -3
    assert dimensionality == 0
    assert labels == [r'$x$ [$\mu$m]', r'$y$ [$\mu$m]', r'$z$ [$\mu$m]']


def test_get_data_1D_x(mocker):
    evc = EvaluationController()
    evc.session.set_file(data_file_path('1D-x.h5'))
    evc.session.set_current_repetition('0')

    evm = evc.session.evaluation_model()

    resolution = evc.session.get_payload_resolution()
    assert resolution == (3, 1, 1)

    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': resolution[0],
        'dim_y': resolution[1],
        'dim_z': resolution[2],
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 2,
    })
    evm.results['brillouin_shift_f'][:] = 1e9  # [Hz]

    data, positions, dimensionality, labels = evc.get_data('brillouin_shift_f')

    positions_x = np.zeros(shape=resolution)
    positions_x[:, 0, 0] = [-1, 0, 1]
    positions_y = np.zeros(shape=resolution)
    positions_y[:] = -2
    positions_z = np.zeros(shape=resolution)
    positions_z[:] = -3

    assert (data == np.ones(shape=resolution)).all()  # [GHz]
    assert (positions[0] == positions_x).all()
    assert (positions[1] == positions_y).all()
    assert (positions[2] == positions_z).all()
    assert dimensionality == 1
    assert labels == [r'$x$ [$\mu$m]', r'$y$ [$\mu$m]', r'$z$ [$\mu$m]']


def test_get_data_1D_y(mocker):
    evc = EvaluationController()
    evc.session.set_file(data_file_path('1D-y.h5'))
    evc.session.set_current_repetition('0')

    evm = evc.session.evaluation_model()

    resolution = evc.session.get_payload_resolution()
    assert resolution == (1, 5, 1)

    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': resolution[0],
        'dim_y': resolution[1],
        'dim_z': resolution[2],
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 2,
    })
    evm.results['brillouin_shift_f'][:] = 1e9  # [Hz]

    data, positions, dimensionality, labels = evc.get_data('brillouin_shift_f')

    positions_x = np.zeros(shape=resolution)
    positions_x[:] = -1
    positions_y = np.zeros(shape=resolution)
    positions_y[0, :, 0] = [-2, -1, 0, 1, 2]
    positions_z = np.zeros(shape=resolution)
    positions_z[:] = -3

    assert (data == np.ones(shape=resolution)).all()  # [GHz]
    assert (positions[0] == positions_x).all()
    assert (positions[1] == positions_y).all()
    assert (positions[2] == positions_z).all()
    assert dimensionality == 1
    assert labels == [r'$x$ [$\mu$m]', r'$y$ [$\mu$m]', r'$z$ [$\mu$m]']


def test_get_data_1D_z(mocker):
    evc = EvaluationController()
    evc.session.set_file(data_file_path('1D-z.h5'))
    evc.session.set_current_repetition('0')

    evm = evc.session.evaluation_model()

    resolution = evc.session.get_payload_resolution()
    assert resolution == (1, 1, 7)

    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': resolution[0],
        'dim_y': resolution[1],
        'dim_z': resolution[2],
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 2,
    })
    evm.results['brillouin_shift_f'][:] = 1e9  # [Hz]

    data, positions, dimensionality, labels = evc.get_data('brillouin_shift_f')

    positions_x = np.zeros(shape=resolution)
    positions_x[:] = -1
    positions_y = np.zeros(shape=resolution)
    positions_y[:] = -2
    positions_z = np.zeros(shape=resolution)
    positions_z[0, 0, :] = [-3, -2, -1, 0, 1, 2, 3]

    assert (data == np.ones(shape=resolution)).all()  # [GHz]
    assert (positions[0] == positions_x).all()
    assert (positions[1] == positions_y).all()
    assert (positions[2] == positions_z).all()
    assert dimensionality == 1
    assert labels == [r'$x$ [$\mu$m]', r'$y$ [$\mu$m]', r'$z$ [$\mu$m]']


def test_get_data_2D_xy(mocker):
    evc = EvaluationController()
    evc.session.set_file(data_file_path('2D-xy.h5'))
    evc.session.set_current_repetition('0')

    evm = evc.session.evaluation_model()

    resolution = evc.session.get_payload_resolution()
    assert resolution == (3, 5, 1)

    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': resolution[0],
        'dim_y': resolution[1],
        'dim_z': resolution[2],
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 2,
    })
    evm.results['brillouin_shift_f'][:] = 1e9  # [Hz]

    data, positions, dimensionality, labels = evc.get_data('brillouin_shift_f')

    data_expected = np.ones(shape=resolution)
    (pos_y, pos_x, pos_z) = np.meshgrid([-2, -1, 0, 1, 2], [-1, 0, 1], [-3])

    assert (data == data_expected).all()
    assert (positions[0] == pos_x).all()
    assert (positions[1] == pos_y).all()
    assert (positions[2] == pos_z).all()
    assert dimensionality == 2
    assert labels == [r'$x$ [$\mu$m]', r'$y$ [$\mu$m]', r'$z$ [$\mu$m]']


def test_get_data_2D_xz(mocker):
    evc = EvaluationController()
    evc.session.set_file(data_file_path('2D-xz.h5'))
    evc.session.set_current_repetition('0')

    evm = evc.session.evaluation_model()

    resolution = evc.session.get_payload_resolution()
    assert resolution == (3, 1, 7)

    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': resolution[0],
        'dim_y': resolution[1],
        'dim_z': resolution[2],
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 2,
    })
    evm.results['brillouin_shift_f'][:] = 1e9  # [Hz]

    data, positions, dimensionality, labels = evc.get_data('brillouin_shift_f')

    data_expected = np.ones(shape=resolution)
    (pos_y, pos_x, pos_z) = np.meshgrid(
        [-2], [-1, 0, 1], [-3, -2, -1, 0, 1, 2, 3]
    )

    assert (data == data_expected).all()
    assert (positions[0] == pos_x).all()
    assert (positions[1] == pos_y).all()
    assert (positions[2] == pos_z).all()
    assert dimensionality == 2
    assert labels == [r'$x$ [$\mu$m]', r'$y$ [$\mu$m]', r'$z$ [$\mu$m]']


def test_get_data_2D_yz(mocker):
    evc = EvaluationController()
    evc.session.set_file(data_file_path('2D-yz.h5'))
    evc.session.set_current_repetition('0')

    evm = evc.session.evaluation_model()

    resolution = evc.session.get_payload_resolution()
    assert resolution == (1, 5, 7)

    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': resolution[0],
        'dim_y': resolution[1],
        'dim_z': resolution[2],
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 2,
    })
    evm.results['brillouin_shift_f'][:] = 1e9  # [Hz]

    data, positions, dimensionality, labels = evc.get_data('brillouin_shift_f')

    data_expected = np.ones(shape=resolution)
    (pos_y, pos_x, pos_z) = np.meshgrid(
        [-2, -1, 0, 1, 2], [-1], [-3, -2, -1, 0, 1, 2, 3]
    )

    assert (data == data_expected).all()
    assert (positions[0] == pos_x).all()
    assert (positions[1] == pos_y).all()
    assert (positions[2] == pos_z).all()
    assert dimensionality == 2
    assert labels == [r'$x$ [$\mu$m]', r'$y$ [$\mu$m]', r'$z$ [$\mu$m]']


def test_get_data_3D(mocker):
    evc = EvaluationController()
    evc.session.set_file(data_file_path('3D.h5'))
    evc.session.set_current_repetition('0')

    evm = evc.session.evaluation_model()

    resolution = evc.session.get_payload_resolution()
    assert resolution == (3, 5, 7)

    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': resolution[0],
        'dim_y': resolution[1],
        'dim_z': resolution[2],
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 1,
        'nr_rayleigh_regions': 2,
    })
    evm.results['brillouin_shift_f'][:] = 1e9  # [Hz]

    data, positions, dimensionality, labels = evc.get_data('brillouin_shift_f')

    data_expected = np.ones(resolution)
    (pos_y, pos_x, pos_z) = np.meshgrid(
        [-2, -1, 0, 1, 2],
        [-1, 0, 1],
        [-3, -2, -1, 0, 1, 2, 3]
    )

    assert dimensionality == 3
    assert (data == data_expected).all()  # [GHz]
    assert (positions[0] == pos_x).all()
    assert (positions[1] == pos_y).all()
    assert (positions[2] == pos_z).all()
    assert labels == [r'$x$ [$\mu$m]', r'$y$ [$\mu$m]', r'$z$ [$\mu$m]']


def test_get_data_3D_peak_index():
    evc = EvaluationController()
    evc.session.set_file(data_file_path('3D.h5'))
    evc.session.set_current_repetition('0')

    evm = evc.session.evaluation_model()

    resolution = evc.session.get_payload_resolution()

    # Initialize results array
    evm.initialize_results_arrays({
        'dim_x': resolution[0],
        'dim_y': resolution[1],
        'dim_z': resolution[2],
        'nr_images': 2,
        'nr_brillouin_regions': 2,
        'nr_brillouin_peaks': 2,
        'nr_rayleigh_regions': 2,
    })
    evm.results['brillouin_shift_f'][:, :, :, :, :, 0] = 2e9  # [Hz]
    evm.results['brillouin_shift_f'][:, :, :, :, :, 1] = 4e9  # [Hz]
    evm.results['brillouin_shift_f'][:, :, :, :, :, 2] = 6e9  # [Hz]

    evm.results['brillouin_peak_fwhm'][:, :, :, :, :, 0] = 1e9  # [Hz]
    evm.results['brillouin_peak_fwhm'][:, :, :, :, :, 1] = 1e9  # [Hz]
    evm.results['brillouin_peak_fwhm'][:, :, :, :, :, 2] = 2e9  # [Hz]

    evm.results['brillouin_peak_intensity'][:, :, :, :, :, 0] = 1e9  # [Hz]
    evm.results['brillouin_peak_intensity'][:, :, :, :, :, 1] = 2e9  # [Hz]
    evm.results['brillouin_peak_intensity'][:, :, :, :, :, 2] = 3e9  # [Hz]

    # Get first peak
    data, positions, dimensionality, labels =\
        evc.get_data('brillouin_shift_f')
    assert data.shape == resolution
    assert (data == 2 * np.ones(resolution)).all()  # [GHz]

    data, positions, dimensionality, labels =\
        evc.get_data('brillouin_shift_f', 0)
    assert data.shape == resolution
    assert (data == 2 * np.ones(resolution)).all()  # [GHz]

    # Get first peak of multi-peak fit
    data, positions, dimensionality, labels =\
        evc.get_data('brillouin_shift_f', 1)
    assert data.shape == resolution
    assert (data == 4 * np.ones(resolution)).all()  # [GHz]

    # Get second peak of multi-peak fit
    data, positions, dimensionality, labels =\
        evc.get_data('brillouin_shift_f', 2)
    assert data.shape == resolution
    assert (data == 6 * np.ones(resolution)).all()  # [GHz]

    # Get average of multi-peak fits
    data, positions, dimensionality, labels =\
        evc.get_data('brillouin_shift_f', 3)
    assert data.shape == resolution
    assert (data == 5 * np.ones(resolution)).all()  # [GHz]

    # Get weighted average of multi-peak fits
    data, positions, dimensionality, labels =\
        evc.get_data('brillouin_shift_f', 4)
    assert data.shape == resolution
    assert (data == 5.5 * np.ones(resolution)).all()  # [GHz]

    # Get single-peak fit when out of bounds
    data, positions, dimensionality, labels =\
        evc.get_data('brillouin_shift_f', 5)
    assert data.shape == resolution
    assert (data == 2 * np.ones(resolution)).all()  # [GHz]


def test_create_bounds(mocker):
    cm = CalibrationModel()
    evm = EvaluationModel()
    mocker.patch('bmlab.session.Session.calibration_model', return_value=cm)
    mocker.patch('bmlab.session.Session.evaluation_model', return_value=evm)
    evc = EvaluationController()

    # Set the calibration data
    cm.set_frequencies('0', 0,
                       list(1e9 * np.linspace(
                           -1, 16, 681).reshape(1, -1)))
    cm.set_frequencies('1', 1,
                       list(1e9 * np.linspace(
                           -0.975, 16.025, 681).reshape(1, -1)))
    cm.set_frequencies('2', 2,
                       list(1e9 * np.linspace(
                           -0.95, 16.05, 681).reshape(1, -1)))

    # If we don't supply regions and times, we get no bounds
    fit_bounds = EvaluationController().create_bounds(None, None, None)
    assert fit_bounds is None

    evm.bounds = [['min', '5'], ['5.5', 'Inf'], ['-inf', 'max']]

    brillouin_regions = [(200, 280), (400, 480)]
    times = [0, 1, 2]
    rayleigh_peaks = np.array([[40, 40, 41], [640, 640, 641]])
    fit_bounds = evc.create_bounds(brillouin_regions, times, rayleigh_peaks)

    np.testing.assert_allclose([
        [
            [[200, 240], [260, np.inf], [-np.inf, 280]],
            [[200, 240], [260, np.inf], [-np.inf, 280]],
            [[200, 241], [261, np.inf], [-np.inf, 280]]
        ], [
            [[440, 480], [-np.inf, 420], [400, np.inf]],
            [[440, 480], [-np.inf, 420], [400, np.inf]],
            [[441, 480], [-np.inf, 421], [400, np.inf]]
        ]
    ], fit_bounds, atol=1.E-3)



from bmlab.models.regions import regions_merge_add_region


def test_merge_add_region():
    regions = [(1, 5), (6, 7)]

    regions_merge_add_region(regions, (8, 9))

    assert regions == [(1, 5), (6, 7), (8, 9)]

    regions_merge_add_region(regions, (8, 10))

    assert regions == [(1, 5), (6, 7), (8, 10)]
